// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

using Smdn.Formats;
using Smdn.Formats.Mime;
using Smdn.IO;

using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMessageInfo {
    private const ImapMessagePartInfo EntireMessageSpecifier = null;

    [Obsolete("use GetMessagePart()")]
    public IImapBodyStructure GetStructureOf(params int[] subSections)
    {
      if (subSections == null || subSections.Length == 0)
        return BodyStructure;
      else
        return ImapBodyStructureUtils.FindSection(BodyStructure, subSections);
    }

    [Obsolete("use GetMessagePart()")]
    public TBodyStructure GetStructureOf<TBodyStructure>(params int[] subSections)
      where TBodyStructure : IImapBodyStructure
    {
      return (TBodyStructure)GetStructureOf(subSections);
    }

    [Obsolete("use GetMessagePart()")]
    public IImapBodyStructure GetStructureOf(string section)
    {
      section = ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(section);

      if (string.IsNullOrEmpty(section))
        return BodyStructure;
      else
        return ImapBodyStructureUtils.FindSection(BodyStructure, section);
    }

    [Obsolete("use GetMessagePart()")]
    public TBodyStructure GetStructureOf<TBodyStructure>(string section)
      where TBodyStructure : IImapBodyStructure
    {
      return (TBodyStructure)GetStructureOf(section);
    }

    /*
     * utility methods for Open*(), Read*()
     */
    [Obsolete]
    private ImapMessagePartInfo ToMessagePart(IImapBodyStructure structure)
    {
      if (staticAttribute == null /* not fetched yed */ ||
          !object.ReferenceEquals(staticAttribute.BodyStructure, structure.GetRootStructure()))
        throw new ArgumentException("invalid structure", "structure");

      return FindMessagePart(delegate(ImapMessagePartInfo part) {
        return part.BodyStructure == structure;
      });
    }

    private static readonly Encoding iso8859_1 = Encoding.GetEncoding("ISO-8859-1");

    /*
     * OpenRead()
     */
    [Obsolete("use ImapMessagePartInfo.OpenRead()")]
    public Stream OpenRead(IImapBodyStructure section)
    {
      return OpenReadImpl(ToMessagePart(section),
                          null,
                          ImapMessageFetchBodyOptions.Default);
    }

    [Obsolete("use ImapMessagePartInfo.OpenRead()")]
    public Stream OpenRead(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      return OpenReadImpl(ToMessagePart(section),
                          null,
                          options);
    }

    public Stream OpenRead()
    {
      return OpenReadImpl(EntireMessageSpecifier,
                          null,
                          ImapMessageFetchBodyOptions.Default);
    }

    public Stream OpenRead(ImapMessageFetchBodyOptions options)
    {
      return OpenReadImpl(EntireMessageSpecifier,
                          null,
                          options);
    }

    public Stream OpenRead(ImapPartialRange? fetchRange)
    {
      return OpenReadImpl(EntireMessageSpecifier,
                          fetchRange,
                          ImapMessageFetchBodyOptions.Default);
    }

    public Stream OpenRead(ImapPartialRange? fetchRange,
                           ImapMessageFetchBodyOptions options)
    {
      return OpenReadImpl(EntireMessageSpecifier,
                          fetchRange,
                          options);
    }

    internal Stream OpenReadImpl(ImapMessagePartInfo messagePart,
                                 ImapPartialRange? fetchRange,
                                 ImapMessageFetchBodyOptions options)
    {
      return FetchMessageBody(messagePart,
                              fetchRange,
                              options,
                              null,
                              (stream, charset) => stream);
    }

    [Obsolete("use ImapMessagePartInfo.OpenRead()", true)]
    public Stream OpenRead(string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.OpenRead()", true)]
    public Stream OpenRead(string section, ImapPartialRange? range)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.OpenRead()", true)]
    public Stream OpenRead(string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.OpenRead()", true)]
    public Stream OpenRead(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    /*
     * ReadAs<TResult>(Converter<Stream>)
     */
    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()")]
    public TResult ReadAs<TResult>(IImapBodyStructure section, Converter<Stream, TResult> converter)
    {
      return ReadAsImpl(ToMessagePart(section),
                        ImapMessageFetchBodyOptions.Default,
                        converter);
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()")]
    public TResult ReadAs<TResult>(IImapBodyStructure section, ImapMessageFetchBodyOptions options, Converter<Stream, TResult> converter)
    {
      return ReadAsImpl(ToMessagePart(section),
                        options,
                        converter);
    }

    public TResult ReadAs<TResult>(Converter<Stream, TResult> converter)
    {
      return ReadAsImpl(EntireMessageSpecifier,
                        ImapMessageFetchBodyOptions.Default,
                        converter);
    }

    public TResult ReadAs<TResult>(ImapMessageFetchBodyOptions options,
                                   Converter<Stream, TResult> converter)
    {
      return ReadAsImpl(EntireMessageSpecifier,
                        options,
                        converter);
    }

    internal TResult ReadAsImpl<TResult>(ImapMessagePartInfo messagePart,
                                         ImapMessageFetchBodyOptions options,
                                         Converter<Stream, TResult> converter)
    {
      if (converter == null)
        throw new ArgumentNullException("converter");

      return FetchMessageBody(messagePart,
                              null,
                              options,
                              null,
                              (stream, charset) => {
        using (stream) {
          return converter(stream);
        }
      });
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()", true)]
    public TResult ReadAs<TResult>(string section, Converter<Stream, TResult> converter)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()", true)]
    public TResult ReadAs<TResult>(string section, ImapMessageFetchBodyOptions options, Converter<Stream, TResult> converter)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()", true)]
    public TResult ReadAs<TResult>(string section, ImapPartialRange? range, Converter<Stream, TResult> converter)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()", true)]
    public TResult ReadAs<TResult>(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options, Converter<Stream, TResult> converter)
    {
      throw new NotImplementedException();
    }

    /*
     * ReadAs<TResult>(Func<Stream, ...>)
     */
    public TResult ReadAs<T, TResult>
      (Func<Stream, T, TResult> read, T arg)
    {
      return ReadAs(ImapMessageFetchBodyOptions.Default, read, arg);
    }

    public TResult ReadAs<T1, T2, TResult>
      (Func<Stream, T1, T2, TResult> read, T1 arg1, T2 arg2)
    {
      return ReadAs(ImapMessageFetchBodyOptions.Default, read, arg1, arg2);
    }

    public TResult ReadAs<T1, T2, T3, TResult>
      (Func<Stream, T1, T2, T3, TResult> read, T1 arg1, T2 arg2, T3 arg3)
    {
      return ReadAs(ImapMessageFetchBodyOptions.Default, read, arg1, arg2, arg3);
    }

    public TResult ReadAs<T1, T2, T3, T4, TResult>
      (Func<Stream, T1, T2, T3, T4, TResult> read, T1 arg1, T2 arg2, T3 arg3, T4 arg4)
    {
      return ReadAs(ImapMessageFetchBodyOptions.Default, read, arg1, arg2, arg3, arg4);
    }

    public TResult ReadAs<T, TResult>
      (ImapMessageFetchBodyOptions options,
       Func<Stream, T, TResult> read, T arg)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(options)) {
        return read(stream, arg);
      }
    }

    public TResult ReadAs<T1, T2, TResult>
      (ImapMessageFetchBodyOptions options,
       Func<Stream, T1, T2, TResult> read, T1 arg1, T2 arg2)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(options)) {
        return read(stream, arg1, arg2);
      }
    }

    public TResult ReadAs<T1, T2, T3, TResult>
      (ImapMessageFetchBodyOptions options,
       Func<Stream, T1, T2, T3, TResult> read, T1 arg1, T2 arg2, T3 arg3)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(options)) {
        return read(stream, arg1, arg2, arg3);
      }
    }

    public TResult ReadAs<T1, T2, T3, T4, TResult>
      (ImapMessageFetchBodyOptions options,
       Func<Stream, T1, T2, T3, T4, TResult> read, T1 arg1, T2 arg2, T3 arg3, T4 arg4)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(options)) {
        return read(stream, arg1, arg2, arg3, arg4);
      }
    }

    /*
     * ReadAllBytes()
     */
    [Obsolete("use ImapMessagePartInfo.ReadAllBytes()")]
    public byte[] ReadAllBytes(IImapBodyStructure section)
    {
      return ReadAllBytesImpl(ToMessagePart(section),
                              ImapMessageFetchBodyOptions.Default);
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllBytes()")]
    public byte[] ReadAllBytes(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      return ReadAllBytesImpl(ToMessagePart(section),
                              options);
    }

    public byte[] ReadAllBytes()
    {
      return ReadAllBytesImpl(EntireMessageSpecifier,
                              ImapMessageFetchBodyOptions.Default);
    }

    public byte[] ReadAllBytes(ImapMessageFetchBodyOptions options)
    {
      return ReadAllBytesImpl(EntireMessageSpecifier,
                              options);
    }

    internal byte[] ReadAllBytesImpl(ImapMessagePartInfo messagePart,
                                     ImapMessageFetchBodyOptions options)
    {
      return FetchMessageBody(messagePart,
                              null,
                              options,
                              null,
                              ReadAllBytesProc);
    }

    private static byte[] ReadAllBytesProc(Stream stream, Encoding charset)
    {
      using (stream) {
        return stream.ReadToEnd();
      }
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllBytes()", true)]
    public byte[] ReadAllBytes(string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllBytes()", true)]
    public byte[] ReadAllBytes(string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllBytes()", true)]
    public byte[] ReadAllBytes(string section, ImapPartialRange? range)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllBytes()", true)]
    public byte[] ReadAllBytes(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    /*
     * WriteTo(Stream)
     */
    [Obsolete("use ImapMessagePartInfo.WriteTo()")]
    public void WriteTo(Stream stream, IImapBodyStructure section)
    {
      WriteToImpl(ToMessagePart(section),
                  ImapMessageFetchBodyOptions.Default,
                  stream);
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()")]
    public void WriteTo(Stream stream, IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      WriteToImpl(ToMessagePart(section),
                  options,
                  stream);
    }

    public void WriteTo(Stream stream)
    {
      WriteToImpl(EntireMessageSpecifier,
                  ImapMessageFetchBodyOptions.Default,
                  stream);
    }

    public void WriteTo(Stream stream,
                        ImapMessageFetchBodyOptions options)
    {
      WriteToImpl(EntireMessageSpecifier,
                  options,
                  stream);
    }

    internal void WriteToImpl(ImapMessagePartInfo messagePart,
                              ImapMessageFetchBodyOptions options,
                              Stream stream)
    {
      if (stream == null)
        throw new ArgumentNullException("stream");
      if (!stream.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream("stream");

      ReadAsImpl(messagePart,
                 options,
                 messageStream => CopyFetchMessageBodyStream(messageStream, stream));
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()", true)]
    public void WriteTo(Stream stream, string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()", true)]
    public void WriteTo(Stream stream, string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()", true)]
    public void WriteTo(Stream stream, string section, ImapPartialRange? range)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()", true)]
    public void WriteTo(Stream stream, string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    private static Stream CopyFetchMessageBodyStream(Stream messageStream, Stream destination)
    {
      var fetchStream = messageStream as ImapFetchMessageBodyStream;

      if (fetchStream == null)
        messageStream.CopyTo(destination);
      else
        fetchStream.ReadToEnd(destination);

      return messageStream;
    }

    private static Stream CopyFetchMessageBodyStream(Stream messageStream, BinaryWriter destination)
    {
      var fetchStream = messageStream as ImapFetchMessageBodyStream;

      if (fetchStream == null)
        messageStream.CopyTo(destination);
      else
        fetchStream.ReadToEnd(destination);

      return messageStream;
    }

    /*
     * WriteTo(BinaryWriter)
     */
    [Obsolete("use ImapMessagePartInfo.WriteTo()")]
    public void WriteTo(BinaryWriter writer, IImapBodyStructure section)
    {
      WriteToImpl(ToMessagePart(section),
                  ImapMessageFetchBodyOptions.Default,
                  writer);
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()")]
    public void WriteTo(BinaryWriter writer, IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      WriteToImpl(ToMessagePart(section),
                  options,
                  writer);
    }

    public void WriteTo(BinaryWriter writer)
    {
      WriteToImpl(EntireMessageSpecifier,
                  ImapMessageFetchBodyOptions.Default,
                  writer);
    }

    public void WriteTo(BinaryWriter writer,
                        ImapMessageFetchBodyOptions options)
    {
      WriteToImpl(EntireMessageSpecifier,
                  options,
                  writer);
    }

    internal void WriteToImpl(ImapMessagePartInfo messagePart,
                              ImapMessageFetchBodyOptions options,
                              BinaryWriter writer)
    {
      if (writer == null)
        throw new ArgumentNullException("writer");

      ReadAsImpl(messagePart,
                 options,
                 messageStream => CopyFetchMessageBodyStream(messageStream, writer));
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()", true)]
    public void WriteTo(BinaryWriter writer, string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()", true)]
    public void WriteTo(BinaryWriter writer, string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()", true)]
    public void WriteTo(BinaryWriter writer, string section, ImapPartialRange? range)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.WriteTo()", true)]
    public void WriteTo(BinaryWriter writer, string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    /*
     * Save()
     */
    [Obsolete("use ImapMessagePartInfo.Save()")]
    public void Save(string path, IImapBodyStructure section)
    {
      SaveImpl(ToMessagePart(section),
               path,
               ImapMessageFetchBodyOptions.Default);
    }

    [Obsolete("use ImapMessagePartInfo.Save()")]
    public void Save(string path, IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      SaveImpl(ToMessagePart(section),
               path,
               options);
    }

    public void Save(string path)
    {
      SaveImpl(EntireMessageSpecifier,
               path,
               ImapMessageFetchBodyOptions.Default);
    }

    public void Save(string path,
                     ImapMessageFetchBodyOptions options)
    {
      SaveImpl(EntireMessageSpecifier,
               path,
               options);
    }

    internal void SaveImpl(ImapMessagePartInfo messagePart,
                           string path,
                           ImapMessageFetchBodyOptions options)
    {
      if (path == null)
        throw new ArgumentNullException("path");
      if (path.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("path");

      ReadAsImpl(messagePart,
                 options,
                 stream => SaveProc(stream, path));
    }

    private static Stream SaveProc(Stream messageStream, string path)
    {
      using (var fileStream = new FileStream(path, FileMode.Create, FileAccess.Write)) {
        return CopyFetchMessageBodyStream(messageStream, fileStream);
      }
    }

    [Obsolete("use ImapMessagePartInfo.Save()", true)]
    public void Save(string path, string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.Save()", true)]
    public void Save(string path, string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.Save()", true)]
    public void Save(string path, string section, ImapPartialRange? range)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.Save()", true)]
    public void Save(string path, string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    /*
     * Download()
     */
    [Obsolete("use ImapMessagePartInfo.Download()")]
    public Stream Download(IImapBodyStructure section)
    {
      return DownloadImpl(ToMessagePart(section),
                          ImapMessageFetchBodyOptions.Default);
    }

    [Obsolete("use ImapMessagePartInfo.Download()")]
    public Stream Download(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      return DownloadImpl(ToMessagePart(section),
                          options);
    }

    public Stream Download()
    {
      return DownloadImpl(EntireMessageSpecifier,
                          ImapMessageFetchBodyOptions.Default);
    }

    public Stream Download(ImapMessageFetchBodyOptions options)
    {
      return DownloadImpl(EntireMessageSpecifier,
                          options);
    }

    internal Stream DownloadImpl(ImapMessagePartInfo messagePart,
                                 ImapMessageFetchBodyOptions options)
    {
      return FetchMessageBody(messagePart,
                              null,
                              options,
                              null,
                              DownloadProc);
    }

    [Obsolete("use ImapMessagePartInfo.Download()", true)]
    public Stream Download(string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.Download()", true)]
    public Stream Download(string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.Download()", true)]
    public Stream Download(string section, ImapPartialRange? range)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.Download()", true)]
    public Stream Download(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    private static Stream DownloadProc(Stream messageStream, Encoding charset)
    {
      using (messageStream) {
        var ret = new ChunkedMemoryStream();

        CopyFetchMessageBodyStream(messageStream, ret);

        ret.Position = 0L;

        return ret;
      }
    }

    /*
     * OpenText
     */
    [Obsolete("use ImapMessagePartInfo.OpenText()")]
    public StreamReader OpenText(IImapBodyStructure section)
    {
      return OpenTextImpl(ToMessagePart(section),
                          iso8859_1,
                          null,
                          ImapMessageFetchBodyOptions.Default);
    }

    [Obsolete("use ImapMessagePartInfo.OpenText()")]
    public StreamReader OpenText(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      return OpenTextImpl(ToMessagePart(section),
                          iso8859_1,
                          null,
                          options);
    }

    public StreamReader OpenText()
    {
      return OpenTextImpl(EntireMessageSpecifier,
                          iso8859_1,
                          null,
                          ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(ImapMessageFetchBodyOptions options)
    {
      return OpenTextImpl(EntireMessageSpecifier,
                          iso8859_1,
                          null,
                          options);
    }

    public StreamReader OpenText(Encoding charset)
    {
      return OpenTextImpl(EntireMessageSpecifier,
                          charset,
                          null,
                          ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(Encoding charset,
                                 ImapMessageFetchBodyOptions options)
    {
      return OpenTextImpl(EntireMessageSpecifier,
                          charset,
                          null,
                          options);
    }

    internal StreamReader OpenTextImpl(ImapMessagePartInfo messagePart,
                                       Encoding charsetForReadingEntireMessage,
                                       EncodingSelectionCallback selectFallbackCharset,
                                       ImapMessageFetchBodyOptions options)
    {
      if (messagePart == EntireMessageSpecifier && charsetForReadingEntireMessage == null)
        throw new ArgumentNullException("charset");

      return FetchMessageBody(messagePart,
                              null,
                              options,
                              selectFallbackCharset,
                              (stream, charset) => new StreamReader(stream, charset ?? charsetForReadingEntireMessage));
    }

    [Obsolete("use ImapMessagePartInfo.OpenText()", true)]
    public StreamReader OpenText(string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.OpenText()", true)]
    public StreamReader OpenText(string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.OpenText()", true)]
    public StreamReader OpenText(string section, Encoding charset)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.OpenText()", true)]
    public StreamReader OpenText(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    /*
     * ReadAs<TResult>(Converter<StreamReader>)
     */
    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()")]
    public TResult ReadAs<TResult>(IImapBodyStructure section, Converter<StreamReader, TResult> converter)
    {
      return ReadAsImpl(ToMessagePart(section),
                        iso8859_1,
                        null,
                        ImapMessageFetchBodyOptions.Default,
                        converter);
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()")]
    public TResult ReadAs<TResult>(IImapBodyStructure section, ImapMessageFetchBodyOptions options, Converter<StreamReader, TResult> converter)
    {
      return ReadAsImpl(ToMessagePart(section),
                        iso8859_1,
                        null,
                        options,
                        converter);
    }

    public TResult ReadAs<TResult>(Converter<StreamReader, TResult> converter)
    {
      return ReadAsImpl(EntireMessageSpecifier,
                        iso8859_1,
                        null,
                        ImapMessageFetchBodyOptions.Default,
                        converter);
    }

    public TResult ReadAs<TResult>(ImapMessageFetchBodyOptions options,
                                   Converter<StreamReader, TResult> converter)
    {
      return ReadAsImpl(EntireMessageSpecifier,
                        iso8859_1,
                        null,
                        options,
                        converter);
    }

    public TResult ReadAs<TResult>(Encoding charset,
                                   Converter<StreamReader, TResult> converter)
    {
      return ReadAsImpl(EntireMessageSpecifier,
                        charset,
                        null,
                        ImapMessageFetchBodyOptions.Default,
                        converter);
    }

    public TResult ReadAs<TResult>(Encoding charset,
                                   ImapMessageFetchBodyOptions options,
                                   Converter<StreamReader, TResult> converter)
    {
      return ReadAsImpl(EntireMessageSpecifier,
                        charset,
                        null,
                        options,
                        converter);
    }

    internal TResult ReadAsImpl<TResult>(ImapMessagePartInfo messagePart,
                                         Encoding charsetForReadingEntireMessage,
                                         EncodingSelectionCallback selectFallbackCharset,
                                         ImapMessageFetchBodyOptions options,
                                         Converter<StreamReader, TResult> converter)
    {
      if (converter == null)
        throw new ArgumentNullException("converter");

      using (var reader = OpenTextImpl(messagePart, charsetForReadingEntireMessage, selectFallbackCharset, options)) {
        return converter(reader);
      }
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()", true)]
    public TResult ReadAs<TResult>(string section, Converter<StreamReader, TResult> converter)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()", true)]
    public TResult ReadAs<TResult>(string section, ImapMessageFetchBodyOptions options, Converter<StreamReader, TResult> converter)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()", true)]
    public TResult ReadAs<TResult>(string section, Encoding charset, Converter<StreamReader, TResult> converter)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAs<TResult>()", true)]
    public TResult ReadAs<TResult>(string section, Encoding charset, ImapMessageFetchBodyOptions options, Converter<StreamReader, TResult> converter)
    {
      throw new NotImplementedException();
    }

    /*
     * ReadLines()
     */
    [Obsolete("use ImapMessagePartInfo.ReadLines()")]
    public IEnumerable<string> ReadLines(IImapBodyStructure structure)
    {
      return ReadLinesImpl(ToMessagePart(structure),
                           iso8859_1,
                           null,
                           ImapMessageFetchBodyOptions.Default);
    }

    [Obsolete("use ImapMessagePartInfo.ReadLines()")]
    public IEnumerable<string> ReadLines(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      return ReadLinesImpl(ToMessagePart(structure),
                           iso8859_1,
                           null,
                           options);
    }

    public IEnumerable<string> ReadLines()
    {
      return ReadLinesImpl(EntireMessageSpecifier,
                           iso8859_1,
                           null,
                           ImapMessageFetchBodyOptions.Default);
    }

    public IEnumerable<string> ReadLines(ImapMessageFetchBodyOptions options)
    {
      return ReadLinesImpl(EntireMessageSpecifier,
                           iso8859_1,
                           null,
                           options);
    }

    public IEnumerable<string> ReadLines(Encoding charset)
    {
      return ReadLinesImpl(EntireMessageSpecifier,
                           charset,
                           null,
                           ImapMessageFetchBodyOptions.Default);
    }

    public IEnumerable<string> ReadLines(Encoding charset,
                                         ImapMessageFetchBodyOptions options)
    {
      return ReadLinesImpl(EntireMessageSpecifier,
                           charset,
                           null,
                           options);
    }

    internal IEnumerable<string> ReadLinesImpl(ImapMessagePartInfo messagePart,
                                               Encoding charsetForReadingEntireMessage,
                                               EncodingSelectionCallback selectFallbackCharset,
                                               ImapMessageFetchBodyOptions options)
    {
      return ReadLinesImpl(OpenTextImpl(messagePart, charsetForReadingEntireMessage, selectFallbackCharset, options));
    }

    private IEnumerable<string> ReadLinesImpl(StreamReader reader)
    {
      using (reader) {
        for (;;) {
          var line = reader.ReadLine();

          if (line == null)
            break;
          else
            yield return line;
        }
      }
    }

    [Obsolete("use ImapMessagePartInfo.ReadLines()", true)]
    public IEnumerable<string> ReadLines(string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadLines()", true)]
    public IEnumerable<string> ReadLines(string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadLines()", true)]
    public IEnumerable<string> ReadLines(string section, Encoding charset)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadLines()", true)]
    public IEnumerable<string> ReadLines(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    /*
     * ReadAllLines()
     */
    [Obsolete("use ImapMessagePartInfo.ReadAllLines()")]
    public string[] ReadAllLines(IImapBodyStructure structure)
    {
      return ReadAllLinesImpl(ToMessagePart(structure),
                              iso8859_1,
                              null,
                              ImapMessageFetchBodyOptions.Default);
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllLines()")]
    public string[] ReadAllLines(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      return ReadAllLinesImpl(ToMessagePart(structure),
                              iso8859_1,
                              null,
                              options);
    }

    public string[] ReadAllLines()
    {
      return ReadAllLinesImpl(EntireMessageSpecifier,
                              iso8859_1,
                              null,
                              ImapMessageFetchBodyOptions.Default);
    }

    public string[] ReadAllLines(ImapMessageFetchBodyOptions options)
    {
      return ReadAllLinesImpl(EntireMessageSpecifier,
                              iso8859_1,
                              null,
                              options);
    }

    public string[] ReadAllLines(Encoding charset)
    {
      return ReadAllLinesImpl(EntireMessageSpecifier,
                              charset,
                              null,
                              ImapMessageFetchBodyOptions.Default);
    }

    public string[] ReadAllLines(Encoding charset,
                                 ImapMessageFetchBodyOptions options)
    {
      return ReadAllLinesImpl(EntireMessageSpecifier,
                              charset,
                              null,
                              options);
    }

    internal string[] ReadAllLinesImpl(ImapMessagePartInfo messagePart,
                                       Encoding charsetForReadingEntireMessage,
                                       EncodingSelectionCallback selectFallbackCharset,
                                       ImapMessageFetchBodyOptions options)
    {
      return ReadAsImpl(messagePart,
                        charsetForReadingEntireMessage,
                        selectFallbackCharset,
                        options,
                        reader => reader.ReadAllLines());
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllLines()", true)]
    public string[] ReadAllLines(string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllLines()", true)]
    public string[] ReadAllLines(string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllLines()", true)]
    public string[] ReadAllLines(string section, Encoding charset)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllLines()", true)]
    public string[] ReadAllLines(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    /*
     * ReadAllText()
     */
    [Obsolete("use ImapMessagePartInfo.ReadAllText()")]
    public string ReadAllText(IImapBodyStructure structure)
    {
      return ReadAllTextImpl(ToMessagePart(structure),
                             iso8859_1,
                             null,
                             ImapMessageFetchBodyOptions.Default);
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllText()")]
    public string ReadAllText(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      return ReadAllTextImpl(ToMessagePart(structure),
                             iso8859_1,
                             null,
                             options);
    }

    public string ReadAllText()
    {
      return ReadAllTextImpl(EntireMessageSpecifier,
                             iso8859_1,
                             null,
                             ImapMessageFetchBodyOptions.Default);
    }

    public string ReadAllText(ImapMessageFetchBodyOptions options)
    {
      return ReadAllTextImpl(EntireMessageSpecifier,
                             iso8859_1,
                             null,
                             options);
    }

    public string ReadAllText(Encoding charset)
    {
      return ReadAllTextImpl(EntireMessageSpecifier,
                             charset,
                             null,
                             ImapMessageFetchBodyOptions.Default);
    }

    public string ReadAllText(Encoding charset,
                              ImapMessageFetchBodyOptions options)
    {
      return ReadAllTextImpl(EntireMessageSpecifier,
                             charset,
                             null,
                             options);
    }

    internal string ReadAllTextImpl(ImapMessagePartInfo messagePart,
                                    Encoding charsetForReadingEntireMessage,
                                    EncodingSelectionCallback selectFallbackCharset,
                                    ImapMessageFetchBodyOptions options)
    {
      return ReadAsImpl(messagePart,
                        charsetForReadingEntireMessage,
                        selectFallbackCharset,
                        options,
                        reader => reader.ReadToEnd());
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllText()", true)]
    public string ReadAllText(string section)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllText()", true)]
    public string ReadAllText(string section, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllText()", true)]
    public string ReadAllText(string section, Encoding charset)
    {
      throw new NotImplementedException();
    }

    [Obsolete("use ImapMessagePartInfo.ReadAllText()", true)]
    public string ReadAllText(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      throw new NotImplementedException();
    }

    /*
     * GetRawHeader(), GetHeader()
     */
    public string GetRawHeader()
    {
      return GetRawHeaderImpl(EntireMessageSpecifier,
                              EmptyArray<string>.Instance);
    }

    public string GetRawHeader(params string[] headerFields)
    {
      return GetRawHeaderImpl(EntireMessageSpecifier,
                              headerFields);
    }

    public string GetRawHeader(/*params*/ IEnumerable<string> headerFields)
    {
      return GetRawHeaderImpl(EntireMessageSpecifier,
                              headerFields);
    }

    internal string GetRawHeaderImpl(ImapMessagePartInfo messagePart,
                                     /*params*/ IEnumerable<string> headerFields)
    {
      return FetchHeaderFields(messagePart,
                               headerFields,
                               ReadRawHeader);
    }

    private static string ReadRawHeader(Stream stream)
    {
      using (var reader = new StreamReader(stream, iso8859_1, false)) {
        return reader.ReadToEnd();
      }
    }

    public KeyValuePair<string, string>[] GetHeader()
    {
      return GetHeaderImpl(EntireMessageSpecifier,
                           false,
                           EmptyArray<string>.Instance);
    }

    public KeyValuePair<string, string>[] GetHeader(params string[] headerFields)
    {
      return GetHeaderImpl(EntireMessageSpecifier,
                           false,
                           headerFields);
    }

    public KeyValuePair<string, string>[] GetHeader(bool decode, params string[] headerFields)
    {
      return GetHeaderImpl(EntireMessageSpecifier,
                           decode,
                           headerFields);
    }

    public KeyValuePair<string, string>[] GetHeader(/*params*/ IEnumerable<string> headerFields)
    {
      return GetHeaderImpl(EntireMessageSpecifier,
                           false,
                           headerFields);
    }

    public KeyValuePair<string, string>[] GetHeader(bool decode, /*params*/ IEnumerable<string> headerFields)
    {
      return GetHeaderImpl(EntireMessageSpecifier,
                           decode,
                           headerFields);
    }

    internal KeyValuePair<string, string>[] GetHeaderImpl(ImapMessagePartInfo messagePart,
                                                          bool decode,
                                                          /*params*/ IEnumerable<string> headerFields)
    {
      return FetchHeaderFields(messagePart,
                               headerFields,
                               stream => DecodeHeaderFields(stream, decode));
    }

    private static KeyValuePair<string, string>[] DecodeHeaderFields(Stream stream, bool decode)
    {
      var headers = Smdn.Formats.Mime.MimeUtils.ParseHeader(stream, true).ToArray();

      for (var i = 0; i < headers.Length; i++) {
        var value = headers[i].Value.Trim();

        if (decode)
          value = MimeEncoding.Decode(value);

        headers[i] = new KeyValuePair<string, string>(headers[i].Key, value);
      }

      return headers;
    }
  }
}

