//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

#if ASYNC_OPS
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using Smdn.Formats;
using Smdn.IO;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMessageInfo {
    /*
     * OpenRead()
     */
    public Stream OpenRead(CancellationToken cancellationToken)
    {
      return FetchMessageBody(ImapMessageFetchBodyOptions.Default,
                              cancellationToken,
                              null);
    }

#if NET_4_5
    public Stream OpenRead(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return FetchMessageBody(ImapMessageFetchBodyOptions.Default,
                              CancellationToken.None,
                              progress);
    }

    public Stream OpenRead(CancellationToken cancellationToken,
                           IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return FetchMessageBody(ImapMessageFetchBodyOptions.Default,
                              cancellationToken,
                              progress);
    }
#endif

    public Stream OpenRead(ImapMessageFetchBodyOptions options,
                           CancellationToken cancellationToken)
    {
      return FetchMessageBody(options,
                              cancellationToken,
                              null);
    }

#if NET_4_5
    public Stream OpenRead(ImapMessageFetchBodyOptions options,
                           IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return FetchMessageBody(options,
                              CancellationToken.None,
                              progress);
    }

    public Stream OpenRead(ImapMessageFetchBodyOptions options,
                           CancellationToken cancellationToken,
                           IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return FetchMessageBody(options,
                              cancellationToken,
                              progress);
    }
#endif

    /*
     * OpenReadAsync()
     */
    public Task<Stream> OpenReadAsync()
    {
      return OpenReadAsyncImpl(EntireMessageSpecifier,
                               ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               null);
    }

    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken)
    {
      return OpenReadAsyncImpl(EntireMessageSpecifier,
                               ImapMessageFetchBodyOptions.Default,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<Stream> OpenReadAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenReadAsyncImpl(EntireMessageSpecifier,
                               ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               progress);
    }

    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenReadAsyncImpl(EntireMessageSpecifier,
                               ImapMessageFetchBodyOptions.Default,
                               cancellationToken,
                               progress);
    }
#endif

    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options)
    {
      return OpenReadAsyncImpl(EntireMessageSpecifier,
                               options,
                               CancellationToken.None,
                               null);
    }

    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken)
    {
      return OpenReadAsyncImpl(EntireMessageSpecifier,
                               options,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenReadAsyncImpl(EntireMessageSpecifier,
                               options,
                               CancellationToken.None,
                               progress);
    }

    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenReadAsyncImpl(EntireMessageSpecifier,
                               options,
                               cancellationToken,
                               progress);
    }
#endif

    internal Task<Stream> OpenReadAsyncImpl(ImapMessagePartInfo messagePart,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
#if NET_4_5
                                            IProgress<ImapMessageFetchProgressInfo> progress
#else
                                            object progress
#endif
                                           )
    {
      return FetchMessageBodyAsync(messagePart,
                                   options,
                                   null,
                                   cancellationToken,
                                   progress,
                                   (stream, charset) => stream);
    }

    /*
     * ReadAsAsync<TResult>(Converter<Stream>)
     */
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             ImapMessageFetchBodyOptions.Default,
                             converter,
                             CancellationToken.None,
                             null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             ImapMessageFetchBodyOptions.Default,
                             converter,
                             cancellationToken,
                             null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             ImapMessageFetchBodyOptions.Default,
                             converter,
                             CancellationToken.None,
                             progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             ImapMessageFetchBodyOptions.Default,
                             converter,
                             cancellationToken,
                             progress);
    }
#endif

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             options,
                             converter,
                             CancellationToken.None,
                             null);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             options,
                             converter,
                             cancellationToken,
                             null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             options,
                             converter,
                             CancellationToken.None,
                             progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             options,
                             converter,
                             cancellationToken,
                             progress);
    }
#endif

    internal Task<TResult> ReadAsAsyncImpl<TResult>(ImapMessagePartInfo messagePart,
                                                    ImapMessageFetchBodyOptions options,
                                                    Converter<Stream, TResult> converter,
                                                    CancellationToken cancellationToken,
#if NET_4_5
                                                    IProgress<ImapMessageFetchProgressInfo> progress
#else
                                                    object progress
#endif
                                                   )
    {
      if (converter == null)
        throw new ArgumentNullException("converter");

      return FetchMessageBodyAsync(messagePart,
                                   options,
                                   null,
                                   cancellationToken,
                                   progress,
                                   (stream, charset) => {
        using (stream) {
          return converter(stream);
        }
      });
    }

    /*
     * ReadAsAsync<TResult>(Func<Stream, ...>)
     */

    /*
     * ReadAllBytesAsync()
     */
    public Task<byte[]> ReadAllBytesAsync()
    {
      return ReadAllBytesAsyncImpl(EntireMessageSpecifier,
                                   ImapMessageFetchBodyOptions.Default,
                                   CancellationToken.None,
                                   null);
    }

    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken)
    {
      return ReadAllBytesAsyncImpl(EntireMessageSpecifier,
                                   ImapMessageFetchBodyOptions.Default,
                                   cancellationToken,
                                   null);
    }

#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllBytesAsyncImpl(EntireMessageSpecifier,
                                   ImapMessageFetchBodyOptions.Default,
                                   CancellationToken.None,
                                   progress);
    }

    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken,
                                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllBytesAsyncImpl(EntireMessageSpecifier,
                                   ImapMessageFetchBodyOptions.Default,
                                   cancellationToken,
                                   progress);
    }
#endif

    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options)
    {
      return ReadAllBytesAsyncImpl(EntireMessageSpecifier,
                                   options,
                                   CancellationToken.None,
                                   null);
    }

    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                          CancellationToken cancellationToken)
    {
      return ReadAllBytesAsyncImpl(EntireMessageSpecifier,
                                   options,
                                   cancellationToken,
                                   null);
    }

#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllBytesAsyncImpl(EntireMessageSpecifier,
                                   options,
                                   CancellationToken.None,
                                   progress);
    }

    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                          CancellationToken cancellationToken,
                                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllBytesAsyncImpl(EntireMessageSpecifier,
                                   options,
                                   cancellationToken,
                                   progress);
    }
#endif

    internal Task<byte[]> ReadAllBytesAsyncImpl(ImapMessagePartInfo messagePart,
                                                ImapMessageFetchBodyOptions options,
                                                CancellationToken cancellationToken,
#if NET_4_5
                                                IProgress<ImapMessageFetchProgressInfo> progress
#else
                                                object progress
#endif
                                               )
    {
      return FetchMessageBodyAsync(messagePart,
                                   options,
                                   null,
                                   cancellationToken,
                                   progress,
                                   ReadAllBytesProc);
    }

    /*
     * WriteToAsync(Stream)
     */
    public Task WriteToAsync(Stream stream)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              stream,
                              ImapMessageFetchBodyOptions.Default,
                              CancellationToken.None,
                              null);
    }

    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              stream,
                              ImapMessageFetchBodyOptions.Default,
                              cancellationToken,
                              null);
    }

#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              stream,
                              ImapMessageFetchBodyOptions.Default,
                              CancellationToken.None,
                              progress);
    }

    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              stream,
                              ImapMessageFetchBodyOptions.Default,
                              cancellationToken,
                              progress);
    }
#endif

    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              stream,
                              options,
                              CancellationToken.None,
                              null);
    }

    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              stream,
                              options,
                              cancellationToken,
                              null);
    }

#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              stream,
                              options,
                              CancellationToken.None,
                              progress);
    }

    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              stream,
                              options,
                              cancellationToken,
                              progress);
    }
#endif

    internal Task WriteToAsyncImpl(ImapMessagePartInfo messagePart,
                                   Stream stream,
                                   ImapMessageFetchBodyOptions options,
                                   CancellationToken cancellationToken,
#if NET_4_5
                                   IProgress<ImapMessageFetchProgressInfo> progress
#else
                                   object progress
#endif
                                  )
    {
      if (stream == null)
        throw new ArgumentNullException("stream");
      if (!stream.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream("stream");

      return ReadAsAsyncImpl(messagePart,
                             options,
                             messageStream => CopyFetchMessageBodyStream(messageStream, stream),
                             cancellationToken,
                             progress);
    }

    /*
     * WriteToAsync(BinaryWriter)
     */
    public Task WriteToAsync(BinaryWriter writer)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              writer,
                              ImapMessageFetchBodyOptions.Default,
                              CancellationToken.None,
                              null);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              writer,
                              ImapMessageFetchBodyOptions.Default,
                              cancellationToken,
                              null);
    }

#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              writer,
                              ImapMessageFetchBodyOptions.Default,
                              CancellationToken.None,
                              progress);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              writer,
                              ImapMessageFetchBodyOptions.Default,
                              cancellationToken,
                              progress);
    }
#endif

    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              writer,
                              options,
                              CancellationToken.None,
                              null);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              writer,
                              options,
                              cancellationToken,
                              null);
    }

#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              writer,
                              options,
                              CancellationToken.None,
                              progress);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return WriteToAsyncImpl(EntireMessageSpecifier,
                              writer,
                              options,
                              cancellationToken,
                              progress);
    }
#endif

    internal Task WriteToAsyncImpl(ImapMessagePartInfo messagePart,
                                   BinaryWriter writer,
                                   ImapMessageFetchBodyOptions options,
                                   CancellationToken cancellationToken,
#if NET_4_5
                                   IProgress<ImapMessageFetchProgressInfo> progress
#else
                                   object progress
#endif
                                  )
    {
      if (writer == null)
        throw new ArgumentNullException("writer");

      return ReadAsAsyncImpl(messagePart,
                             options,
                             messageStream => CopyFetchMessageBodyStream(messageStream, writer),
                             cancellationToken,
                             progress);
    }

    /*
     * SaveAsync()
     */
    public Task SaveAsync(string path)
    {
      return SaveAsyncImpl(EntireMessageSpecifier,
                           path,
                           ImapMessageFetchBodyOptions.Default,
                           CancellationToken.None,
                           null);
    }

    public Task SaveAsync(string path,
                          CancellationToken cancellationToken)
    {
      return SaveAsyncImpl(EntireMessageSpecifier,
                           path,
                           ImapMessageFetchBodyOptions.Default,
                           cancellationToken,
                           null);
    }

#if NET_4_5
    public Task SaveAsync(string path,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return SaveAsyncImpl(EntireMessageSpecifier,
                           path,
                           ImapMessageFetchBodyOptions.Default,
                           CancellationToken.None,
                           progress);
    }

    public Task SaveAsync(string path,
                          CancellationToken cancellationToken,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return SaveAsyncImpl(EntireMessageSpecifier,
                           path,
                           ImapMessageFetchBodyOptions.Default,
                           cancellationToken,
                           progress);
    }
#endif

    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options)
    {
      return SaveAsyncImpl(EntireMessageSpecifier,
                           path,
                           options,
                           CancellationToken.None,
                           null);
    }

    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options,
                          CancellationToken cancellationToken)
    {
      return SaveAsyncImpl(EntireMessageSpecifier,
                           path,
                           options,
                           cancellationToken,
                           null);
    }

#if NET_4_5
    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return SaveAsyncImpl(EntireMessageSpecifier,
                           path,
                           options,
                           CancellationToken.None,
                           progress);
    }

    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options,
                          CancellationToken cancellationToken,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return SaveAsyncImpl(EntireMessageSpecifier,
                           path,
                           options,
                           cancellationToken,
                           progress);
    }
#endif

    internal Task SaveAsyncImpl(ImapMessagePartInfo messagePart,
                                string path,
                                ImapMessageFetchBodyOptions options,
                                CancellationToken cancellationToken,
#if NET_4_5
                                IProgress<ImapMessageFetchProgressInfo> progress
#else
                                object progress
#endif
                               )
    {
      if (path == null)
        throw new ArgumentNullException("path");
      if (path.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("path");

      return ReadAsAsyncImpl(messagePart,
                             options,
                             stream => SaveProc(stream, path),
                             cancellationToken,
                             progress);
    }

    /*
     * DownloadAsync()
     */
    public Task<Stream> DownloadAsync()
    {
      return DownloadAsyncImpl(EntireMessageSpecifier,
                               ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               null);
    }

    public Task<Stream> DownloadAsync(CancellationToken cancellationToken)
    {
      return DownloadAsyncImpl(EntireMessageSpecifier,
                               ImapMessageFetchBodyOptions.Default,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<Stream> DownloadAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return DownloadAsyncImpl(EntireMessageSpecifier,
                               ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               progress);
    }

    public Task<Stream> DownloadAsync(CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return DownloadAsyncImpl(EntireMessageSpecifier,
                               ImapMessageFetchBodyOptions.Default,
                               cancellationToken,
                               progress);
    }
#endif

    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options)
    {
      return DownloadAsyncImpl(EntireMessageSpecifier,
                               options,
                               CancellationToken.None,
                               null);
    }

    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken)
    {
      return DownloadAsyncImpl(EntireMessageSpecifier,
                               options,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return DownloadAsyncImpl(EntireMessageSpecifier,
                               options,
                               CancellationToken.None,
                               progress);
    }

    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return DownloadAsyncImpl(EntireMessageSpecifier,
                               options,
                               CancellationToken.None,
                               progress);
    }
#endif

    // implementation of DownloadAsync
    internal Task<Stream> DownloadAsyncImpl(ImapMessagePartInfo messagePart,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
#if NET_4_5
                                            IProgress<ImapMessageFetchProgressInfo> progress
#else
                                            object progress
#endif
                                           )
    {
      return FetchMessageBodyAsync(messagePart,
                                   options,
                                   null,
                                   cancellationToken,
                                   progress,
                                   DownloadProc);
    }

    /*
     * OpenTextAsync()
     */
    public Task<StreamReader> OpenTextAsync()
    {
      return OpenTextAsyncImpl(EntireMessageSpecifier,
                               iso8859_1,
                               null,
                               ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               null);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset)
    {
      return OpenTextAsyncImpl(EntireMessageSpecifier,
                               charset,
                               null,
                               ImapMessageFetchBodyOptions.Default,
                               CancellationToken.None,
                               null);
    }

    public Task<StreamReader> OpenTextAsync(ImapMessageFetchBodyOptions options)
    {
      return OpenTextAsyncImpl(EntireMessageSpecifier,
                               iso8859_1,
                               null,
                               options,
                               CancellationToken.None,
                               null);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options)
    {
      return OpenTextAsyncImpl(EntireMessageSpecifier,
                               charset,
                               null,
                               options,
                               CancellationToken.None,
                               null);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken)
    {
      return OpenTextAsyncImpl(EntireMessageSpecifier,
                               charset,
                               null,
                               options,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenTextAsyncImpl(EntireMessageSpecifier,
                               charset,
                               null,
                               options,
                               CancellationToken.None,
                               progress);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return OpenTextAsyncImpl(EntireMessageSpecifier,
                               charset,
                               null,
                               options,
                               cancellationToken,
                               progress);
    }
#endif

    internal Task<StreamReader> OpenTextAsyncImpl(ImapMessagePartInfo messagePart,
                                                  Encoding charsetForReadingEntireMessage,
                                                  EncodingSelectionCallback selectFallbackCharset,
                                                  ImapMessageFetchBodyOptions options,
                                                  CancellationToken cancellationToken,
#if NET_4_5
                                                  IProgress<ImapMessageFetchProgressInfo> progress
#else
                                                  object progress
#endif
                                                 )
    {
      if (messagePart == EntireMessageSpecifier && charsetForReadingEntireMessage == null)
        throw new ArgumentNullException("charset");

      return FetchMessageBodyAsync(messagePart,
                                   options,
                                   selectFallbackCharset,
                                   cancellationToken,
                                   progress,
                                   (stream, charset) => new StreamReader(stream, charset ?? charsetForReadingEntireMessage));
    }

    /*
     * ReadAsAsync<TResult>(Converter<StreamReader>)
     */
    public Task<TResult> ReadAsAsync<TResult>(Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             iso8859_1,
                             null,
                             ImapMessageFetchBodyOptions.Default,
                             converter,
                             CancellationToken.None,
                             null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             charset,
                             null,
                             ImapMessageFetchBodyOptions.Default,
                             converter,
                             CancellationToken.None,
                             null);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             iso8859_1,
                             null,
                             options,
                             converter,
                             CancellationToken.None,
                             null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             charset,
                             null,
                             options,
                             converter,
                             CancellationToken.None,
                             null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             charset,
                             null,
                             options,
                             converter,
                             cancellationToken,
                             null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             charset,
                             null,
                             options,
                             converter,
                             CancellationToken.None,
                             progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsyncImpl(EntireMessageSpecifier,
                             charset,
                             null,
                             options,
                             converter,
                             cancellationToken,
                             progress);
    }
#endif

    internal Task<TResult> ReadAsAsyncImpl<TResult>(ImapMessagePartInfo messagePart,
                                                    Encoding charsetForReadingEntireMessage,
                                                    EncodingSelectionCallback selectFallbackCharset,
                                                    ImapMessageFetchBodyOptions options,
                                                    Converter<StreamReader, TResult> converter,
                                                    CancellationToken cancellationToken,
#if NET_4_5
                                                    IProgress<ImapMessageFetchProgressInfo> progress
#else
                                                    object progress
#endif
                                                   )
    {
      if (messagePart == EntireMessageSpecifier && charsetForReadingEntireMessage == null)
        throw new ArgumentNullException("charset");
      if (converter == null)
        throw new ArgumentNullException("converter");

      return FetchMessageBodyAsync(messagePart,
                                   options,
                                   selectFallbackCharset,
                                   cancellationToken,
                                   progress,
                                   (stream, charset) => {
        using (var reader = new StreamReader(stream, charset ?? charsetForReadingEntireMessage)) {
          return converter(reader);
        }
      });
    }

    /*
     * ReadLinesAsync()
     */
#if false
    // implementation of ReadLinesAsync()
    public Task<IEnumerable<string>> ReadLinesAsync(Encoding charset,
                                                    ImapMessageFetchBodyOptions options,
                                                    CancellationToken cancellationToken,
                                                    IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAsAsync(charset,
                         options,
                         reader => reader.ReadLines(), // yield return does not work
                         cancellationToken,
                         progress);
    }
#endif

    /*
     * ReadAllLinesAsync()
     */
    public Task<string[]> ReadAllLinesAsync()
    {
      return ReadAllLinesAsyncImpl(EntireMessageSpecifier,
                                   iso8859_1,
                                   null,
                                   ImapMessageFetchBodyOptions.Default,
                                   CancellationToken.None,
                                   null);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset)
    {
      return ReadAllLinesAsyncImpl(EntireMessageSpecifier,
                                   charset,
                                   null,
                                   ImapMessageFetchBodyOptions.Default,
                                   CancellationToken.None,
                                   null);
    }

    public Task<string[]> ReadAllLinesAsync(ImapMessageFetchBodyOptions options)
    {
      return ReadAllLinesAsyncImpl(EntireMessageSpecifier,
                                   iso8859_1,
                                   null,
                                   options,
                                   CancellationToken.None,
                                   null);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options)
    {
      return ReadAllLinesAsyncImpl(EntireMessageSpecifier,
                                   charset,
                                   null,
                                   options,
                                   CancellationToken.None,
                                   null);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken)
    {
      return ReadAllLinesAsyncImpl(EntireMessageSpecifier,
                                   charset,
                                   null,
                                   options,
                                   cancellationToken,
                                   null);
    }

#if NET_4_5
    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllLinesAsyncImpl(EntireMessageSpecifier,
                                   charset,
                                   null,
                                   options,
                                   CancellationToken.None,
                                   progress);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllLinesAsyncImpl(EntireMessageSpecifier,
                                   charset,
                                   null,
                                   options,
                                   cancellationToken,
                                   progress);
    }
#endif

    internal Task<string[]> ReadAllLinesAsyncImpl(ImapMessagePartInfo messagePart,
                                                  Encoding charset,
                                                  EncodingSelectionCallback selectFallbackCharset,
                                                  ImapMessageFetchBodyOptions options,
                                                  CancellationToken cancellationToken,
#if NET_4_5
                                                  IProgress<ImapMessageFetchProgressInfo> progress
#else
                                                  object progress
#endif
                                                 )
    {
      return ReadAsAsyncImpl(messagePart,
                             charset,
                             selectFallbackCharset,
                             options,
                             reader => reader.ReadAllLines(),
                             cancellationToken,
                             progress);
    }

    /*
     * ReadAllTextAsync()
     */
    public Task<string> ReadAllTextAsync()
    {
      return ReadAllTextAsyncImpl(EntireMessageSpecifier,
                                  iso8859_1,
                                  null,
                                  ImapMessageFetchBodyOptions.Default,
                                  CancellationToken.None,
                                  null);
    }

    public Task<string> ReadAllTextAsync(Encoding charset)
    {
      return ReadAllTextAsyncImpl(EntireMessageSpecifier,
                                  charset,
                                  null,
                                  ImapMessageFetchBodyOptions.Default,
                                  CancellationToken.None,
                                  null);
    }

    public Task<string> ReadAllTextAsync(ImapMessageFetchBodyOptions options)
    {
      return ReadAllTextAsyncImpl(EntireMessageSpecifier,
                                  iso8859_1,
                                  null,
                                  options,
                                  CancellationToken.None,
                                  null);
    }

    public Task<string> ReadAllTextAsync(Encoding charset,
                                         ImapMessageFetchBodyOptions options)
    {
      return ReadAllTextAsyncImpl(EntireMessageSpecifier,
                                  charset,
                                  null,
                                  options,
                                  CancellationToken.None,
                                  null);
    }

    public Task<string> ReadAllTextAsync(Encoding charset,
                                         ImapMessageFetchBodyOptions options,
                                         CancellationToken cancellationToken)
    {
      return ReadAllTextAsyncImpl(EntireMessageSpecifier,
                                  charset,
                                  null,
                                  options,
                                  cancellationToken,
                                  null);
    }

#if NET_4_5
    public Task<string> ReadAllTextAsync(Encoding charset,
                                         ImapMessageFetchBodyOptions options,
                                         IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllTextAsyncImpl(EntireMessageSpecifier,
                                  charset,
                                  null,
                                  options,
                                  CancellationToken.None,
                                  progress);
    }

    public Task<string> ReadAllTextAsync(Encoding charset,
                                         ImapMessageFetchBodyOptions options,
                                         CancellationToken cancellationToken,
                                         IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return ReadAllTextAsyncImpl(EntireMessageSpecifier,
                                  charset,
                                  null,
                                  options,
                                  cancellationToken,
                                  progress);
    }
#endif

    internal Task<string> ReadAllTextAsyncImpl(ImapMessagePartInfo messagePart,
                                               Encoding charset,
                                               EncodingSelectionCallback selectFallbackCharset,
                                               ImapMessageFetchBodyOptions options,
                                               CancellationToken cancellationToken,
#if NET_4_5
                                               IProgress<ImapMessageFetchProgressInfo> progress
#else
                                               object progress
#endif
                                              )
    {
      return ReadAsAsyncImpl(messagePart,
                             charset,
                             selectFallbackCharset,
                             options,
                             reader => reader.ReadToEnd(),
                             cancellationToken,
                             progress);
    }

    /*
     * GetRawHeaderAsync(), GetHeaderAsync()
     */
    public Task<string> GetRawHeaderAsync()
    {
      return GetRawHeaderAsyncImpl(EntireMessageSpecifier,
                                   EmptyArray<string>.Instance);
    }

    public Task<string> GetRawHeaderAsync(params string[] headerFields)
    {
      return GetRawHeaderAsyncImpl(EntireMessageSpecifier,
                                   headerFields);
    }

    public Task<string> GetRawHeaderAsync(/*params*/ IEnumerable<string> headerFields)
    {
      return GetRawHeaderAsyncImpl(EntireMessageSpecifier,
                                   headerFields);
    }

    internal Task<string> GetRawHeaderAsyncImpl(ImapMessagePartInfo messagePart,
                                                /*params*/ IEnumerable<string> headerFields)
    {
      return FetchHeaderFieldsAsync(messagePart,
                                    headerFields,
                                    ReadRawHeader);
    }

    public Task<KeyValuePair<string, string>[]> GetHeaderAsync()
    {
      return GetHeaderAsyncImpl(EntireMessageSpecifier,
                                false,
                                EmptyArray<string>.Instance);
    }

    public Task<KeyValuePair<string, string>[]> GetHeaderAsync(params string[] headerFields)
    {
      return GetHeaderAsyncImpl(EntireMessageSpecifier,
                                false,
                                headerFields);
    }

    public Task<KeyValuePair<string, string>[]> GetHeaderAsync(bool decode,
                                                               params string[] headerFields)
    {
      return GetHeaderAsyncImpl(EntireMessageSpecifier,
                                decode,
                                headerFields);
    }

    public Task<KeyValuePair<string, string>[]> GetHeaderAsync(/*params*/ IEnumerable<string> headerFields)
    {
      return GetHeaderAsyncImpl(EntireMessageSpecifier,
                                false,
                                headerFields);
    }

    public Task<KeyValuePair<string, string>[]> GetHeaderAsync(bool decode,
                                                               /*params*/ IEnumerable<string> headerFields)
    {
      return GetHeaderAsyncImpl(EntireMessageSpecifier,
                                decode,
                                headerFields);
    }

    internal Task<KeyValuePair<string, string>[]> GetHeaderAsyncImpl(ImapMessagePartInfo messagePart,
                                                                     bool decode,
                                                                     /*params*/ IEnumerable<string> headerFields)
    {
      return FetchHeaderFieldsAsync(messagePart,
                                    headerFields,
                                    stream => DecodeHeaderFields(stream, decode));
    }
  }
}
#endif
