﻿//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

#if ASYNC_OPS
using System.Threading;
using System.Threading.Tasks;
#endif

using Smdn.Formats;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMessagePartInfo {
    /*
     * OpenRead()
     */
    public Stream OpenRead()
    {
      return message.OpenReadImpl(this,
                                  (ImapPartialRange?)null,
                                  ImapMessageFetchBodyOptions.Default);
    }

    public Stream OpenRead(ImapMessageFetchBodyOptions options)
    {
      return message.OpenReadImpl(this,
                                  (ImapPartialRange?)null,
                                  options);
    }

    public Stream OpenRead(ImapPartialRange? fetchRange)
    {
      return message.OpenReadImpl(this,
                                  fetchRange,
                                  ImapMessageFetchBodyOptions.Default);
    }

    public Stream OpenRead(ImapPartialRange? fetchRange,
                           ImapMessageFetchBodyOptions options)
    {
      return message.OpenReadImpl(this,
                                  fetchRange,
                                  options);
    }

    /*
     * ReadAs<TResult>(Converter<Stream>)
     */
    public TResult ReadAs<TResult>(Converter<Stream, TResult> converter)
    {
      return message.ReadAsImpl(this,
                                ImapMessageFetchBodyOptions.Default,
                                converter);
    }

    public TResult ReadAs<TResult>(ImapMessageFetchBodyOptions options,
                                   Converter<Stream, TResult> converter)
    {
      return message.ReadAsImpl(this,
                                options,
                                converter);
    }

    /*
     * ReadAllBytes()
     */
    public byte[] ReadAllBytes()
    {
      return message.ReadAllBytesImpl(this,
                                      ImapMessageFetchBodyOptions.Default);
    }

    public byte[] ReadAllBytes(ImapMessageFetchBodyOptions options)
    {
      return message.ReadAllBytesImpl(this,
                                      options);
    }

    /*
     * WriteTo(Stream)
     */
    public void WriteTo(Stream stream)
    {
      message.WriteToImpl(this,
                          ImapMessageFetchBodyOptions.Default,
                          stream);
    }

    public void WriteTo(Stream stream,
                        ImapMessageFetchBodyOptions options)
    {
      message.WriteToImpl(this,
                          options,
                          stream);
    }

    /*
     * WriteTo(BinaryWriter)
     */
    public void WriteTo(BinaryWriter writer)
    {
      message.WriteToImpl(this,
                          ImapMessageFetchBodyOptions.Default,
                          writer);
    }

    public void WriteTo(BinaryWriter writer,
                        ImapMessageFetchBodyOptions options)
    {
      message.WriteToImpl(this,
                          options,
                          writer);
    }

    /*
     * Save()
     */
    public void Save(string path)
    {
      message.SaveImpl(this,
                       path,
                       ImapMessageFetchBodyOptions.Default);
    }

    public void Save(string path,
                     ImapMessageFetchBodyOptions options)
    {
      message.SaveImpl(this,
                       path,
                       options);
    }

    /*
     * Download()
     */
    public Stream Download()
    {
      return message.DownloadImpl(this,
                                  ImapMessageFetchBodyOptions.Default);
    }

    public Stream Download(ImapMessageFetchBodyOptions options)
    {
      return message.DownloadImpl(this,
                                  options);
    }

    /*
     * OpenText
     */
    public StreamReader OpenText()
    {
      return message.OpenTextImpl(this,
                                  null,
                                  null,
                                  ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(ImapMessageFetchBodyOptions options)
    {
      return message.OpenTextImpl(this,
                                  null,
                                  null,
                                  options);
    }

    public StreamReader OpenText(ImapMessageFetchBodyOptions options,
                                 EncodingSelectionCallback selectFallbackCharset)
    {
      return message.OpenTextImpl(this,
                                  null,
                                  selectFallbackCharset,
                                  options);
    }

    /*
     * ReadAs<TResult>(Converter<StreamReader>)
     */
    public TResult ReadAs<TResult>(Converter<StreamReader, TResult> converter)
    {
      return message.ReadAsImpl(this,
                                null,
                                null,
                                ImapMessageFetchBodyOptions.Default,
                                converter);
    }

    public TResult ReadAs<TResult>(ImapMessageFetchBodyOptions options,
                                   Converter<StreamReader, TResult> converter)
    {
      return message.ReadAsImpl(this,
                                null,
                                null,
                                options,
                                converter);
    }

    public TResult ReadAs<TResult>(ImapMessageFetchBodyOptions options,
                                   EncodingSelectionCallback selectFallbackCharset,
                                   Converter<StreamReader, TResult> converter)
    {
      return message.ReadAsImpl(this,
                                null,
                                selectFallbackCharset,
                                options,
                                converter);
    }

    /*
     * ReadLines()
     */
    public IEnumerable<string> ReadLines()
    {
      return message.ReadLinesImpl(this,
                                   null,
                                   null,
                                   ImapMessageFetchBodyOptions.Default);
    }

    public IEnumerable<string> ReadLines(ImapMessageFetchBodyOptions options)
    {
      return message.ReadLinesImpl(this,
                                   null,
                                   null,
                                   options);
    }

    public IEnumerable<string> ReadLines(ImapMessageFetchBodyOptions options,
                                         EncodingSelectionCallback selectFallbackCharset)
    {
      return message.ReadLinesImpl(this,
                                   null,
                                   selectFallbackCharset,
                                   options);
    }

    /*
     * ReadAllLines()
     */
    public string[] ReadAllLines()
    {
      return message.ReadAllLinesImpl(this,
                                      null,
                                      null,
                                      ImapMessageFetchBodyOptions.Default);
    }

    public string[] ReadAllLines(ImapMessageFetchBodyOptions options)
    {
      return message.ReadAllLinesImpl(this,
                                      null,
                                      null,
                                      options);
    }

    public string[] ReadAllLines(ImapMessageFetchBodyOptions options,
                                 EncodingSelectionCallback selectFallbackCharset)
    {
      return message.ReadAllLinesImpl(this,
                                      null,
                                      selectFallbackCharset,
                                      options);
    }

    /*
     * ReadAllText()
     */
    public string ReadAllText()
    {
      return message.ReadAllTextImpl(this,
                                     null,
                                     null,
                                     ImapMessageFetchBodyOptions.Default);
    }

    public string ReadAllText(ImapMessageFetchBodyOptions options)
    {
      return message.ReadAllTextImpl(this,
                                     null,
                                     null,
                                     options);
    }

    public string ReadAllText(ImapMessageFetchBodyOptions options,
                              EncodingSelectionCallback selectFallbackCharset)
    {
      return message.ReadAllTextImpl(this,
                                     null,
                                     selectFallbackCharset,
                                     options);
    }

#if ASYNC_OPS
    /*
     * OpenReadAsync()
     */
    public Task<Stream> OpenReadAsync()
    {
      return message.OpenReadAsyncImpl(this,
                                       ImapMessageFetchBodyOptions.Default,
                                       CancellationToken.None,
                                       null);
    }

    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken)
    {
      return message.OpenReadAsyncImpl(this,
                                       ImapMessageFetchBodyOptions.Default,
                                       cancellationToken,
                                       null);
    }

#if NET_4_5
    public Task<Stream> OpenReadAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.OpenReadAsyncImpl(this,
                                       ImapMessageFetchBodyOptions.Default,
                                       CancellationToken.None,
                                       progress);
    }

    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.OpenReadAsyncImpl(this,
                                       ImapMessageFetchBodyOptions.Default,
                                       cancellationToken,
                                       progress);
    }
#endif

    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options)
    {
      return message.OpenReadAsyncImpl(this,
                                       options,
                                       CancellationToken.None,
                                       null);
    }

    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken)
    {
      return message.OpenReadAsyncImpl(this,
                                       options,
                                       cancellationToken,
                                       null);
    }

#if NET_4_5
    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.OpenReadAsyncImpl(this,
                                       options,
                                       CancellationToken.None,
                                       progress);
    }

    public Task<Stream> OpenReadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.OpenReadAsyncImpl(this,
                                       options,
                                       cancellationToken,
                                       progress);
    }
#endif

    /*
     * ReadAsAsync<TResult>(Converter<Stream>)
     */
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter)
    {
      return message.ReadAsAsyncImpl(this,
                                     ImapMessageFetchBodyOptions.Default,
                                     converter,
                                     CancellationToken.None,
                                     null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return message.ReadAsAsyncImpl(this,
                                     ImapMessageFetchBodyOptions.Default,
                                     converter,
                                     cancellationToken,
                                     null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAsAsyncImpl(this,
                                     ImapMessageFetchBodyOptions.Default,
                                     converter,
                                     CancellationToken.None,
                                     progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAsAsyncImpl(this,
                                     ImapMessageFetchBodyOptions.Default,
                                     converter,
                                     cancellationToken,
                                     progress);
    }
#endif

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter)
    {
      return message.ReadAsAsyncImpl(this,
                                     options,
                                     converter,
                                     CancellationToken.None,
                                     null);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return message.ReadAsAsyncImpl(this,
                                     options,
                                     converter,
                                     cancellationToken,
                                     null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAsAsyncImpl(this,
                                     options,
                                     converter,
                                     CancellationToken.None,
                                     progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAsAsyncImpl(this,
                                     options,
                                     converter,
                                     cancellationToken,
                                     progress);
    }
#endif

    /*
     * ReadAllBytesAsync()
     */
    public Task<byte[]> ReadAllBytesAsync()
    {
      return message.ReadAllBytesAsyncImpl(this,
                                           ImapMessageFetchBodyOptions.Default,
                                           CancellationToken.None,
                                           null);
    }

    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken)
    {
      return message.ReadAllBytesAsyncImpl(this,
                                           ImapMessageFetchBodyOptions.Default,
                                           cancellationToken,
                                           null);
    }

#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllBytesAsyncImpl(this,
                                           ImapMessageFetchBodyOptions.Default,
                                           CancellationToken.None,
                                           progress);
    }

    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken,
                                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllBytesAsyncImpl(this,
                                           ImapMessageFetchBodyOptions.Default,
                                           cancellationToken,
                                           progress);
    }
#endif

    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options)
    {
      return message.ReadAllBytesAsyncImpl(this,
                                           options,
                                           CancellationToken.None,
                                           null);
    }

    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                          CancellationToken cancellationToken)
    {
      return message.ReadAllBytesAsyncImpl(this,
                                           options,
                                           cancellationToken,
                                           null);
    }

#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllBytesAsyncImpl(this,
                                           options,
                                           CancellationToken.None,
                                           progress);
    }

    public Task<byte[]> ReadAllBytesAsync(ImapMessageFetchBodyOptions options,
                                          CancellationToken cancellationToken,
                                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllBytesAsyncImpl(this,
                                           options,
                                           cancellationToken,
                                           progress);
    }
#endif

    /*
     * WriteToAsync(Stream)
     */
    public Task WriteToAsync(Stream stream)
    {
      return message.WriteToAsyncImpl(this,
                                      stream,
                                      ImapMessageFetchBodyOptions.Default,
                                      CancellationToken.None,
                                      null);
    }

    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken)
    {
      return message.WriteToAsyncImpl(this,
                                      stream,
                                      ImapMessageFetchBodyOptions.Default,
                                      cancellationToken,
                                      null);
    }

#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.WriteToAsyncImpl(this,
                                      stream,
                                      ImapMessageFetchBodyOptions.Default,
                                      CancellationToken.None,
                                      progress);
    }

    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.WriteToAsyncImpl(this,
                                      stream,
                                      ImapMessageFetchBodyOptions.Default,
                                      cancellationToken,
                                      progress);
    }
#endif

    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options)
    {
      return message.WriteToAsyncImpl(this,
                                      stream,
                                      options,
                                      CancellationToken.None,
                                      null);
    }

    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken)
    {
      return message.WriteToAsyncImpl(this,
                                      stream,
                                      options,
                                      cancellationToken,
                                      null);
    }

#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.WriteToAsyncImpl(this,
                                      stream,
                                      options,
                                      CancellationToken.None,
                                      progress);
    }

    public Task WriteToAsync(Stream stream,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.WriteToAsyncImpl(this,
                                      stream,
                                      options,
                                      cancellationToken,
                                      progress);
    }
#endif

    /*
     * WriteToAsync(BinaryWriter)
     */
    public Task WriteToAsync(BinaryWriter writer)
    {
      return message.WriteToAsyncImpl(this,
                                      writer,
                                      ImapMessageFetchBodyOptions.Default,
                                      CancellationToken.None,
                                      null);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken)
    {
      return message.WriteToAsyncImpl(this,
                                      writer,
                                      ImapMessageFetchBodyOptions.Default,
                                      cancellationToken,
                                      null);
    }

#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.WriteToAsyncImpl(this,
                                      writer,
                                      ImapMessageFetchBodyOptions.Default,
                                      CancellationToken.None,
                                      progress);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.WriteToAsyncImpl(this,
                                      writer,
                                      ImapMessageFetchBodyOptions.Default,
                                      cancellationToken,
                                      progress);
    }
#endif

    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options)
    {
      return message.WriteToAsyncImpl(this,
                                      writer,
                                      options,
                                      CancellationToken.None,
                                      null);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken)
    {
      return message.WriteToAsyncImpl(this,
                                      writer,
                                      options,
                                      cancellationToken,
                                      null);
    }

#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.WriteToAsyncImpl(this,
                                      writer,
                                      options,
                                      CancellationToken.None,
                                      progress);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             ImapMessageFetchBodyOptions options,
                             CancellationToken cancellationToken,
                             IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.WriteToAsyncImpl(this,
                                      writer,
                                      options,
                                      cancellationToken,
                                      progress);
    }
#endif

    /*
     * SaveAsync()
     */
    public Task SaveAsync(string path)
    {
      return message.SaveAsyncImpl(this,
                                   path,
                                   ImapMessageFetchBodyOptions.Default,
                                   CancellationToken.None,
                                   null);
    }

    public Task SaveAsync(string path,
                          CancellationToken cancellationToken)
    {
      return message.SaveAsyncImpl(this,
                                   path,
                                   ImapMessageFetchBodyOptions.Default,
                                   cancellationToken,
                                   null);
    }

#if NET_4_5
    public Task SaveAsync(string path,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.SaveAsyncImpl(this,
                                   path,
                                   ImapMessageFetchBodyOptions.Default,
                                   CancellationToken.None,
                                   progress);
    }

    public Task SaveAsync(string path,
                          CancellationToken cancellationToken,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.SaveAsyncImpl(this,
                                   path,
                                   ImapMessageFetchBodyOptions.Default,
                                   cancellationToken,
                                   progress);
    }
#endif

    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options)
    {
      return message.SaveAsyncImpl(this,
                                   path,
                                   options,
                                   CancellationToken.None,
                                   null);
    }

    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options,
                          CancellationToken cancellationToken)
    {
      return message.SaveAsyncImpl(this,
                                   path,
                                   options,
                                   cancellationToken,
                                   null);
    }

#if NET_4_5
    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.SaveAsyncImpl(this,
                                   path,
                                   options,
                                   CancellationToken.None,
                                   progress);
    }

    public Task SaveAsync(string path,
                          ImapMessageFetchBodyOptions options,
                          CancellationToken cancellationToken,
                          IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.SaveAsyncImpl(this,
                                   path,
                                   options,
                                   cancellationToken,
                                   progress);
    }
#endif

    /*
     * DownloadAsync()
     */
    public Task<Stream> DownloadAsync()
    {
      return message.DownloadAsyncImpl(this,
                                       ImapMessageFetchBodyOptions.Default,
                                       CancellationToken.None,
                                       null);
    }

    public Task<Stream> DownloadAsync(CancellationToken cancellationToken)
    {
      return message.DownloadAsyncImpl(this,
                                       ImapMessageFetchBodyOptions.Default,
                                       cancellationToken,
                                       null);
    }

#if NET_4_5
    public Task<Stream> DownloadAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.DownloadAsyncImpl(this,
                                       ImapMessageFetchBodyOptions.Default,
                                       CancellationToken.None,
                                       progress);
    }

    public Task<Stream> DownloadAsync(CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.DownloadAsyncImpl(this,
                                       ImapMessageFetchBodyOptions.Default,
                                       cancellationToken,
                                       progress);
    }
#endif

    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options)
    {
      return message.DownloadAsyncImpl(this,
                                       options,
                                       CancellationToken.None,
                                       null);
    }

    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken)
    {
      return message.DownloadAsyncImpl(this,
                                       options,
                                       cancellationToken,
                                       null);
    }

#if NET_4_5
    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.DownloadAsyncImpl(this,
                                       options,
                                       CancellationToken.None,
                                       progress);
    }

    public Task<Stream> DownloadAsync(ImapMessageFetchBodyOptions options,
                                      CancellationToken cancellationToken,
                                      IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.DownloadAsyncImpl(this,
                                       options,
                                       CancellationToken.None,
                                       progress);
    }
#endif

    /*
     * OpenTextAsync()
     */
    public Task<StreamReader> OpenTextAsync()
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       null,
                                       ImapMessageFetchBodyOptions.Default,
                                       CancellationToken.None,
                                       null);
    }

    public Task<StreamReader> OpenTextAsync(CancellationToken cancellationToken)
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       null,
                                       ImapMessageFetchBodyOptions.Default,
                                       cancellationToken,
                                       null);
    }

#if NET_4_5
    public Task<StreamReader> OpenTextAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       null,
                                       ImapMessageFetchBodyOptions.Default,
                                       CancellationToken.None,
                                       progress);
    }

    public Task<StreamReader> OpenTextAsync(CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       null,
                                       ImapMessageFetchBodyOptions.Default,
                                       cancellationToken,
                                       progress);
    }
#endif

    public Task<StreamReader> OpenTextAsync(ImapMessageFetchBodyOptions options)
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       null,
                                       options,
                                       CancellationToken.None,
                                       null);
    }

    public Task<StreamReader> OpenTextAsync(ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken)
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       null,
                                       options,
                                       cancellationToken,
                                       null);
    }

    public Task<StreamReader> OpenTextAsync(ImapMessageFetchBodyOptions options,
                                            EncodingSelectionCallback selectFallbackCharset)
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       selectFallbackCharset,
                                       options,
                                       CancellationToken.None,
                                       null);
    }

#if NET_4_5
    public Task<StreamReader> OpenTextAsync(ImapMessageFetchBodyOptions options,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       null,
                                       options,
                                       CancellationToken.None,
                                       progress);
    }

    public Task<StreamReader> OpenTextAsync(ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       null,
                                       options,
                                       cancellationToken,
                                       progress);
    }

    public Task<StreamReader> OpenTextAsync(ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress,
                                            EncodingSelectionCallback selectFallbackCharset)
    {
      return message.OpenTextAsyncImpl(this,
                                       null,
                                       selectFallbackCharset,
                                       options,
                                       cancellationToken,
                                       progress);
    }
#endif

    /*
     * ReadAsAsync<TResult>(Converter<StreamReader>)
     */
    public Task<TResult> ReadAsAsync<TResult>(Converter<StreamReader, TResult> converter)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     null,
                                     ImapMessageFetchBodyOptions.Default,
                                     converter,
                                     CancellationToken.None,
                                     null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     null,
                                     ImapMessageFetchBodyOptions.Default,
                                     converter,
                                     cancellationToken,
                                     null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Converter<StreamReader, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     null,
                                     ImapMessageFetchBodyOptions.Default,
                                     converter,
                                     CancellationToken.None,
                                     progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     null,
                                     ImapMessageFetchBodyOptions.Default,
                                     converter,
                                     cancellationToken,
                                     progress);
    }
#endif

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     null,
                                     options,
                                     converter,
                                     CancellationToken.None,
                                     null);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     null,
                                     options,
                                     converter,
                                     cancellationToken,
                                     null);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              EncodingSelectionCallback selectFallbackCharset)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     selectFallbackCharset,
                                     options,
                                     converter,
                                     CancellationToken.None,
                                     null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     null,
                                     options,
                                     converter,
                                     CancellationToken.None,
                                     progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     null,
                                     options,
                                     converter,
                                     cancellationToken,
                                     progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(ImapMessageFetchBodyOptions options,
                                              EncodingSelectionCallback selectFallbackCharset,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAsAsyncImpl(this,
                                     null,
                                     selectFallbackCharset,
                                     options,
                                     converter,
                                     cancellationToken,
                                     progress);
    }
#endif

    /*
     * ReadAllLinesAsync()
     */
    public Task<string[]> ReadAllLinesAsync()
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           null,
                                           ImapMessageFetchBodyOptions.Default,
                                           CancellationToken.None,
                                           null);
    }

    public Task<string[]> ReadAllLinesAsync(CancellationToken cancellationToken)
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           null,
                                           ImapMessageFetchBodyOptions.Default,
                                           cancellationToken,
                                           null);
    }

#if NET_4_5
    public Task<string[]> ReadAllLinesAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           null,
                                           ImapMessageFetchBodyOptions.Default,
                                           CancellationToken.None,
                                           progress);
    }

    public Task<string[]> ReadAllLinesAsync(CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           null,
                                           ImapMessageFetchBodyOptions.Default,
                                           cancellationToken,
                                           progress);
    }
#endif

    public Task<string[]> ReadAllLinesAsync(ImapMessageFetchBodyOptions options)
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           null,
                                           options,
                                           CancellationToken.None,
                                           null);
    }

    public Task<string[]> ReadAllLinesAsync(ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken)
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           null,
                                           options,
                                           cancellationToken,
                                           null);
    }

    public Task<string[]> ReadAllLinesAsync(ImapMessageFetchBodyOptions options,
                                            EncodingSelectionCallback selectFallbackCharset)
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           selectFallbackCharset,
                                           options,
                                           CancellationToken.None,
                                           null);
    }

#if NET_4_5
    public Task<string[]> ReadAllLinesAsync(ImapMessageFetchBodyOptions options,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           null,
                                           options,
                                           CancellationToken.None,
                                           progress);
    }

    public Task<string[]> ReadAllLinesAsync(ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           null,
                                           options,
                                           cancellationToken,
                                           progress);
    }

    public Task<string[]> ReadAllLinesAsync(ImapMessageFetchBodyOptions options,
                                            CancellationToken cancellationToken,
                                            IProgress<ImapMessageFetchProgressInfo> progress,
                                            EncodingSelectionCallback selectFallbackCharset)
    {
      return message.ReadAllLinesAsyncImpl(this,
                                           null,
                                           selectFallbackCharset,
                                           options,
                                           cancellationToken,
                                           progress);
    }
#endif

    /*
     * ReadAllTextAsync()
     */
    public Task<string> ReadAllTextAsync()
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          null,
                                          ImapMessageFetchBodyOptions.Default,
                                          CancellationToken.None,
                                          null);
    }

    public Task<string> ReadAllTextAsync(CancellationToken cancellationToken)
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          null,
                                          ImapMessageFetchBodyOptions.Default,
                                          cancellationToken,
                                          null);
    }

#if NET_4_5
    public Task<string> ReadAllTextAsync(IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          null,
                                          ImapMessageFetchBodyOptions.Default,
                                          CancellationToken.None,
                                          progress);
    }

    public Task<string> ReadAllTextAsync(CancellationToken cancellationToken,
                                         IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          null,
                                          ImapMessageFetchBodyOptions.Default,
                                          cancellationToken,
                                          progress);
    }
#endif

    public Task<string> ReadAllTextAsync(ImapMessageFetchBodyOptions options)
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          null,
                                          options,
                                          CancellationToken.None,
                                          null);
    }

    public Task<string> ReadAllTextAsync(ImapMessageFetchBodyOptions options,
                                         CancellationToken cancellationToken)
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          null,
                                          options,
                                          cancellationToken,
                                          null);
    }

    public Task<string> ReadAllTextAsync(ImapMessageFetchBodyOptions options,
                                         EncodingSelectionCallback selectFallbackCharset)
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          selectFallbackCharset,
                                          options,
                                          CancellationToken.None,
                                          null);
    }
#if NET_4_5
    public Task<string> ReadAllTextAsync(ImapMessageFetchBodyOptions options,
                                         IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          null,
                                          options,
                                          CancellationToken.None,
                                          progress);
    }

    public Task<string> ReadAllTextAsync(ImapMessageFetchBodyOptions options,
                                         CancellationToken cancellationToken,
                                         IProgress<ImapMessageFetchProgressInfo> progress)
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          null,
                                          options,
                                          cancellationToken,
                                          progress);
    }

    public Task<string> ReadAllTextAsync(ImapMessageFetchBodyOptions options,
                                         CancellationToken cancellationToken,
                                         IProgress<ImapMessageFetchProgressInfo> progress,
                                         EncodingSelectionCallback selectFallbackCharset)
    {
      return message.ReadAllTextAsyncImpl(this,
                                          null,
                                          selectFallbackCharset,
                                          options,
                                          cancellationToken,
                                          progress);
    }
#endif

#endif // ASYNC_OPS

    public string GetRawHeader()
    {
      return message.GetRawHeaderImpl(this, null);
    }

    public KeyValuePair<string, string>[] GetHeader()
    {
      return message.GetHeaderImpl(this, false, EmptyArray<string>.Instance);
    }

    public KeyValuePair<string, string>[] GetHeader(bool decode)
    {
      return message.GetHeaderImpl(this, decode, EmptyArray<string>.Instance);
    }

#if ASYNC_OPS
    public Task<string> GetRawHeaderAsync()
    {
      return message.GetRawHeaderAsyncImpl(this, null);
    }

    public Task<KeyValuePair<string, string>[]> GetHeaderAsync()
    {
      return message.GetHeaderAsyncImpl(this, false, EmptyArray<string>.Instance);
    }

    public Task<KeyValuePair<string, string>[]> GetHeaderAsync(bool decode)
    {
      return message.GetHeaderAsyncImpl(this, decode, EmptyArray<string>.Instance);
    }
#endif // ASYNC_OPS
  }
}

