#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.Text;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapOpenedMailboxInfoOperationsFetchingTests {
    [Test]
    public void TestGetMessages()
    {
      var selectResp =
        "* 5 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages();

        Assert.AreEqual(5L, mailbox.ExistMessageCount);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "* 5 EXPUNGE\r\n" +
                                       "* 4 EXISTS\r\n" + 
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH 1:4 (UID)\r\n",
                                       "* FETCH 1 (UID 3)\r\n" +
                                       "* FETCH 2 (UID 5)\r\n" +
                                       "* FETCH 3 (UID 7)\r\n" +
                                       "* FETCH 4 (UID 9)\r\n" +
                                       "$tag OK done\r\n");
        });

        var enumeratedMessages = new ImapMessageInfo[4];

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(messages.Client.IsBusy);

          Assert.IsTrue(enumerator.MoveNext());

          Assert.AreEqual(4L, mailbox.ExistMessageCount);

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          enumeratedMessages[0] = enumerator.Current;

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(5L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");

          enumeratedMessages[1] = enumerator.Current;

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(7L, enumerator.Current.Uid, "#2 uid");
          Assert.AreEqual(3L, enumerator.Current.Sequence, "#2 seq");

          enumeratedMessages[2] = enumerator.Current;

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#3 non-null");
          Assert.AreEqual(9L, enumerator.Current.Uid, "#3 uid");
          Assert.AreEqual(4L, enumerator.Current.Sequence, "#3 seq");

          enumeratedMessages[3] = enumerator.Current;

          Assert.IsFalse(enumerator.MoveNext());
        }

        /*
         * re-enumerate
         */
        Assert.AreEqual(4L, mailbox.ExistMessageCount);

        responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "* 2 EXPUNGE\r\n" +
                                       "* 1 EXPUNGE\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH 1:2 (UID)\r\n",
                                       "* FETCH 1 (UID 7)\r\n" +
                                       "* FETCH 2 (UID 9)\r\n" +
                                       "$tag OK done\r\n");
        });

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(messages.Client.IsBusy);

          Assert.IsTrue(enumerator.MoveNext());

          Assert.AreEqual(2L, mailbox.ExistMessageCount);

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(7L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.AreSame(enumeratedMessages[2], enumerator.Current, "returns same instance #0");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(9L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");

          Assert.AreSame(enumeratedMessages[3], enumerator.Current, "returns same instance #1");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessages_NoMessageExists()
    {
      var selectResp =
        "* 2 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "* 1 EXPUNGE\r\n" +
                                       "* 1 EXPUNGE\r\n" +
                                       "$tag OK done\r\n");
        });

        var messages = mailbox.GetMessages().ToList();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(0L, mailbox.ExistMessageCount);

        Assert.AreEqual(0, messages.Count);
      });
    }

    [Test]
    public void TestGetMessageByUid()
    {
      GetMessageByUidOrSequence(true, mailbox => mailbox.GetMessageByUid(3L));
    }

    [Test]
    public void TestGetMessageBySequence()
    {
      GetMessageByUidOrSequence(false, mailbox => mailbox.GetMessageBySequence(1L));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetMessageByUidAsync()
    {
      GetMessageByUidOrSequence(true, mailbox => mailbox.GetMessageByUidAsync(3L).Result);
    }

    [Test]
    public void TestGetMessageBySequenceAsync()
    {
      GetMessageByUidOrSequence(false, mailbox => mailbox.GetMessageBySequenceAsync(1L).Result);
    }
#endif

    private void GetMessageByUidOrSequence(bool uid, Func<ImapOpenedMailboxInfo, ImapMessageInfo> testFunc)
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse(uid ? "UID FETCH 3 (UID)\r\n" : "FETCH 1 (UID)\r\n",
                                       "* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");
        });

        var message = testFunc(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(message);
        Assert.AreEqual(3L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(23L, message.UidValidity);
        Assert.AreSame(mailbox, message.Mailbox);

        responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse(uid ? "UID FETCH 3 (UID)\r\n" : "FETCH 1 (UID)\r\n",
                                       "* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");
        });

        var messageSecond = testFunc(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreSame(message, messageSecond, "returns same instance");
      });
    }

    [Test]
    public void TestGetMessageByUid_NotFound()
    {
      GetMessageByUid_NotFound(mailbox => mailbox.GetMessageByUid(3L));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetMessageByUidAsync_NotFound()
    {
      GetMessageByUid_NotFound(mailbox => mailbox.GetMessageByUidAsync(3L).Result);
    }
#endif

    private void GetMessageByUid_NotFound(Func<ImapOpenedMailboxInfo, ImapMessageInfo> testFunc)
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("UID FETCH 3 (UID)\r\n",
                                       "$tag OK done\r\n");
        });

        var ex = TestUtils.Throws<ImapMessageNotFoundException>(() => testFunc(mailbox));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.SequenceOrUidSet);
        Assert.AreEqual("3", ex.SequenceOrUidSet.ToString());

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMessageNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.SequenceOrUidSet);
          Assert.AreEqual("3", deserialized.SequenceOrUidSet.ToString());
        });
      });
    }

    [Test]
    public void TestGetMessageByUid_ArgumentOutOfRange()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageByUid(0L));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetMessageByUidAsync_ArgumentOutOfRange()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageByUidAsync(0L));
      });
    }
#endif

    [Test]
    public void TestGetMessageBySequence_ArgumentOutOfRange()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageBySequence(0L));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetMessageBySequenceAsync_ArgumentOutOfRange()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageBySequenceAsync(0L));
      });
    }
#endif

    [Test]
    public void TestGetMessageBySequence_GreaterThanExistCount_1()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageBySequence(1L));
      });
    }

    [Test]
    public void TestGetMessageBySequence_GreaterThanExistCount_2()
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageBySequence(2L));
      });
    }

    [Test]
    public void TestGetMessageBySequence_NoResponse()
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                       "$tag NO such message\r\n");
        });

        Assert.Throws<ImapErrorResponseException>(() => mailbox.GetMessageBySequence(1L));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetMessageFirstUnseen()
    {
      GetMessageFirstUnseen(mailbox => mailbox.GetMessageFirstUnseen());
    }

#if ASYNC_OPS
    [Test]
    public void TestGetMessageFirstUnseenAsync()
    {
      GetMessageFirstUnseen(mailbox => mailbox.GetMessageFirstUnseenAsync().Result);
    }
#endif

    private void GetMessageFirstUnseen(Func<ImapOpenedMailboxInfo, ImapMessageInfo> testFunc)
    {
      var selectResp =
        "* 18 EXISTS\r\n" +
        "* OK [UNSEEN 17]\r\n" +
        "* OK [UIDVALIDITY 3857529045]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(17L, mailbox.FirstUnseenMessageNumber);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("FETCH 17 (UID)\r\n",
                                       "* FETCH 17 (UID 20)\r\n" +
                                       "$tag OK done\r\n");
        });

        var message = testFunc(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(message);
        Assert.AreEqual(20L, message.Uid);
        Assert.AreEqual(17L, message.Sequence);
        Assert.AreEqual(3857529045L, message.UidValidity);
        Assert.AreSame(mailbox, message.Mailbox);
      });
    }

    [Test]
    public void TestGetMessageFirstUnseen_FirstUnseenMissing()
    {
      GetMessageFirstUnseen_FirstUnseenMissing(mailbox => mailbox.GetMessageFirstUnseen());
    }

#if ASYNC_OPS
    [Test]
    public void TestGetMessageFirstUnseenAsync_FirstUnseenMissing()
    {
      GetMessageFirstUnseen_FirstUnseenMissing(mailbox => mailbox.GetMessageFirstUnseenAsync().Result);
    }
#endif

    private void GetMessageFirstUnseen_FirstUnseenMissing(Func<ImapOpenedMailboxInfo, ImapMessageInfo> testFunc)
    {
      var selectResp =
        "* 18 EXISTS\r\n" +
        "* OK [UIDVALIDITY 3857529045]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.AreEqual(0L, mailbox.FirstUnseenMessageNumber);

        Assert.IsNull(testFunc(mailbox));
      });
    }

    [Test]
    public void TestGetMessages_ByUids()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(1L, 3L, 5L);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("FETCH 1,3,5 (UID)\r\n",
                                       "* FETCH 1 (UID 1)\r\n" +
                                       "* FETCH 3 (UID 3)\r\n" +
                                       "* FETCH 5 (UID 5)\r\n" +
                                       "$tag OK done\r\n");
        });

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(1L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(3L, enumerator.Current.Sequence, "#1 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(5L, enumerator.Current.Uid, "#2 uid");
          Assert.AreEqual(5L, enumerator.Current.Sequence, "#2 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
          "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp, GetMessagesWithSearchCriteria);
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresDisabled()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.Searchres);

        GetMessagesWithSearchCriteria(server, mailbox);
      });
    }

    private void GetMessagesWithSearchCriteria(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox)
    {
      var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

      var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
        server.EnqueueTaggedResponse("SEARCH RECENT\r\n",
                                     "* SEARCH 1 2 3\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                     "* FETCH 1 (UID 3)\r\n" +
                                     "* FETCH 2 (UID 5)\r\n" +
                                     "* FETCH 3 (UID 7)\r\n" +
                                     "$tag OK done\r\n");
      });

      using (var enumerator = messages.GetEnumerator()) {
        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(enumerator.MoveNext());

        Assert.IsNotNull(enumerator.Current, "#0 non-null");
        Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
        Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

        Assert.IsTrue(enumerator.MoveNext());

        Assert.IsNotNull(enumerator.Current, "#1 non-null");
        Assert.AreEqual(5L, enumerator.Current.Uid, "#1 uid");
        Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");

        Assert.IsTrue(enumerator.MoveNext());

        Assert.IsNotNull(enumerator.Current, "#2 non-null");
        Assert.AreEqual(7L, enumerator.Current.Uid, "#2 uid");
        Assert.AreEqual(3L, enumerator.Current.Sequence, "#2 seq");

        Assert.IsFalse(enumerator.MoveNext());
      }

      /*
       * re-enumerate
       */
      responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
        server.EnqueueTaggedResponse("SEARCH RECENT\r\n",
                                     "* SEARCH 4\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("FETCH 4 (UID)\r\n",
                                     "* FETCH 4 (UID 8)\r\n" +
                                     "$tag OK done\r\n");
      });

      using (var enumerator = messages.GetEnumerator()) {
        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(enumerator.MoveNext());

        Assert.IsNotNull(enumerator.Current, "#0 non-null");
        Assert.AreEqual(8L, enumerator.Current.Uid, "#0 uid");
        Assert.AreEqual(4L, enumerator.Current.Sequence, "#0 seq");

        Assert.IsFalse(enumerator.MoveNext());
      }
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Subject("test"), Encoding.UTF8);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SEARCH CHARSET utf-8 SUBJECT {4}\r\ntest\r\n",
                                       "$tag NO [BADCHARSET (US-ASCII Shift_JIS)] charset not supported\r\n");
        });

        var ex = Assert.Throws<Smdn.Formats.EncodingNotSupportedException>(() => messages.GetEnumerator());

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("utf-8", ex.EncodingName);
        StringAssert.Contains("US-ASCII", ex.Message);
        StringAssert.Contains("Shift_JIS", ex.Message);
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresCapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) RECENT\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH $ (UID)\r\n",
                                       "* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");
        });

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresCapable_NoNotSaved()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) RECENT\r\n",
                                       "$tag NO [NOTSAVED] not saved\r\n");
          server.EnqueueTaggedResponse("SEARCH RECENT\r\n",
                                       "* SEARCH 1\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                       "* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");
        });

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresCapable_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Subject("test"),
                                           Encoding.GetEncoding("Shift_JIS"));

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SEARCH RETURN (SAVE) CHARSET shift_jis SUBJECT {4}\r\ntest\r\n",
                                       "$tag NO [BADCHARSET] charset not supported\r\n");
        });

        var ex = Assert.Throws<Smdn.Formats.EncodingNotSupportedException>(() => messages.GetEnumerator());

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("shift_jis", ex.EncodingName);
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_NothingMatched()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SEARCH RECENT\r\n",
                                       "* SEARCH\r\n" +
                                       "$tag OK done\r\n");
        });

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessagesUnseen()
    {
      GetMessagesWithPresetSearchCriteria("UNSEEN", mailbox => mailbox.GetMessagesUnseen());
    }

    [Test]
    public void TestGetMessagesRecent()
    {
      GetMessagesWithPresetSearchCriteria("RECENT", mailbox => mailbox.GetMessagesRecent());
    }

    [Test]
    public void TestGetMessagesNew()
    {
      GetMessagesWithPresetSearchCriteria("NEW", mailbox => mailbox.GetMessagesNew());
    }

    private void GetMessagesWithPresetSearchCriteria(string expectedSearchCriteria,
                                                     Func<ImapOpenedMailboxInfo, ImapMessageInfoList> getMessages)
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SEARCH " + expectedSearchCriteria + "\r\n",
                                       "* SEARCH 1 2 3\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                       "* FETCH 1 (UID 3)\r\n" +
                                       "* FETCH 2 (UID 5)\r\n" +
                                       "* FETCH 3 (UID 7)\r\n" +
                                       "$tag OK done\r\n");
        });

        var messages = getMessages(mailbox).ToArray();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(3, messages.Length);
      });
    }

    [Test]
    public void TestGetSortedMessages()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Sort},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Arrival, ImapSearchCriteria.Recent);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SORT (ARRIVAL) utf-8 RECENT\r\n",
                                       "* SORT 2 4 1\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH 2,4,1 (UID)\r\n",
                                       "* FETCH 1 (UID 2)\r\n" +
                                       "* FETCH 2 (UID 4)\r\n" +
                                       "* FETCH 4 (UID 8)\r\n" +
                                       "$tag OK done\r\n");
        });

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#0 seq");
          Assert.AreEqual(4L, enumerator.Current.Uid, "#0 uid");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(4L, enumerator.Current.Sequence, "#1 seq");
          Assert.AreEqual(8L, enumerator.Current.Uid, "#1 uid");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#2 seq");
          Assert.AreEqual(2L, enumerator.Current.Uid, "#2 uid");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetSortedMessages_SortIncapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetSortedMessages(ImapSortCriteria.Subject, ImapSearchCriteria.Recent));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetSortedMessages_SortCriteriaIncapable_SortDisplay()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Sort}, "INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetSortedMessages(ImapSortCriteria.DisplayTo, ImapSearchCriteria.Recent));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetSortedMessages_NothingMatched()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Sort},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Arrival, ImapSearchCriteria.Recent);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SORT (ARRIVAL) utf-8 RECENT\r\n",
                                       "* SORT\r\n" +
                                       "$tag OK done\r\n");
        });

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetThreadedMessages()
    {
      GetThreadedMessages((mailbox, algorithm) => mailbox.GetThreadedMessages(algorithm));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetThreadedMessagesAsync()
    {
      GetThreadedMessages((mailbox, algorithm) => mailbox.GetThreadedMessagesAsync(algorithm).Result);
    }
#endif

    private void GetThreadedMessages(Func<ImapOpenedMailboxInfo, ImapThreadingAlgorithm, ImapMessageInfoThread> testFunc)
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.ThreadReferences},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("THREAD REFERENCES utf-8 ALL\r\n",
                                       "* THREAD (2)(3 6 (4 23)(44))\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH 2,3,6,4,23,44 (UID)\r\n",
                                       "* FETCH 2 (UID 2)\r\n" +
                                       "* FETCH 3 (UID 3)\r\n" +
                                       "* FETCH 6 (UID 6)\r\n" +
                                       "* FETCH 4 (UID 4)\r\n" +
                                       "* FETCH 23 (UID 23)\r\n" +
                                       "* FETCH 44 (UID 44)\r\n" +
                                       "$tag OK done\r\n");
        });

        var thread = testFunc(mailbox, ImapThreadingAlgorithm.References);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(thread);
        Assert.IsTrue(thread.IsRoot);
        Assert.IsNull(thread.Parent);
        Assert.AreEqual(0, thread.Depth);
        Assert.Throws<NotSupportedException>(() => Assert.IsNull(thread.Message));

        /*
          -- 2
          -- 3
             `-- 6
                 |-- 4
                 |   `-- 23
                 |
                 `-- 44
        */
        var expectedMessageSequences = new[] {
          2L, 3L, 6L, 4L, 23L, 44L,
        };
        var expectedDepths = new[] {
          1L, 1L, 2L, 3L, 4L, 3L,
        };
        var expectedChildrenCounts = new[] {
          0L, 1L, 2L, 1L, 0L, 0L,
        };
        var traversedMessageCount = 0;

        thread.Traverse(delegate(ImapMessageInfoThread t) {
          Assert.IsFalse(t.IsRoot, "IsRoot #{0}", traversedMessageCount);
          Assert.IsNotNull(t.Parent, "Parent #{0}", traversedMessageCount);
          Assert.AreEqual(expectedDepths[traversedMessageCount], t.Depth, "Depth #{0}", traversedMessageCount);
          Assert.IsNotNull(t.Children, "Children #{0}", traversedMessageCount);
          Assert.AreEqual(expectedChildrenCounts[traversedMessageCount], t.Children.Count, "Children.Count #{0}", traversedMessageCount);
          Assert.AreEqual(expectedMessageSequences[traversedMessageCount], t.Message.Sequence, "Message.Sequence #{0}", traversedMessageCount);

          traversedMessageCount++;
        });

        Assert.AreEqual(6, traversedMessageCount);
      });
    }

    [Test]
    public void TestGetThreadedMessages_ThreadingAlgorithmIncapable_References()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetThreadedMessages(ImapThreadingAlgorithm.References));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetThreadedMessagesAsync_ThreadingAlgorithmIncapable_References()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetThreadedMessagesAsync(ImapThreadingAlgorithm.References));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestGetThreadedMessages_ThreadingAlgorithmIncapable_OrderedSubject()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX", selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetThreadedMessages(ImapThreadingAlgorithm.OrderedSubject));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetThreadedMessages_NothingMatched()
    {
      GetThreadedMessages_NothingMatched((mailbox, algorithm) => mailbox.GetThreadedMessages(algorithm));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetThreadedMessagesAsync_NothingMatched()
    {
      GetThreadedMessages_NothingMatched((mailbox, algorithm) => mailbox.GetThreadedMessagesAsync(algorithm).Result);
    }
#endif

    private void GetThreadedMessages_NothingMatched(Func<ImapOpenedMailboxInfo, ImapThreadingAlgorithm, ImapMessageInfoThread> testFunc)
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.ThreadReferences},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("THREAD REFERENCES utf-8 ALL\r\n",
                                       "* THREAD\r\n" +
                                       "$tag OK done\r\n");
        });

        var thread = testFunc(mailbox, ImapThreadingAlgorithm.References);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(thread);
        Assert.IsTrue(thread.IsRoot);
        Assert.IsNull(thread.Parent);
        Assert.AreEqual(0, thread.Depth);
        Assert.Throws<NotSupportedException>(() => Assert.IsNull(thread.Message));
        Assert.IsEmpty(thread.Children);

        thread.Traverse(delegate(ImapMessageInfoThread t) {
          Assert.Fail("children traversed");
        });
      });
    }

    [Test]
    public void TestGetThreadedMessages_NoBadCharset()
    {
      GetThreadedMessages_NoBadCharset((mailbox, algorithm, criteria, encoding) => mailbox.GetThreadedMessages(algorithm, criteria, encoding));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetThreadedMessagesAsync_NoBadCharset()
    {
      GetThreadedMessages_NoBadCharset((mailbox, algorithm, criteria, encoding) => mailbox.GetThreadedMessagesAsync(algorithm, criteria, encoding).Result);
    }
#endif

    private void GetThreadedMessages_NoBadCharset(Func<ImapOpenedMailboxInfo, ImapThreadingAlgorithm, ImapSearchCriteria, Encoding, ImapMessageInfoThread> testFunc)
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.ThreadReferences},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("THREAD REFERENCES shift_jis SUBJECT {4}\r\ntest\r\n",
                                       "$tag NO [BADCHARSET] charset not supported\r\n");
        });

        var ex = TestUtils.Throws<Smdn.Formats.EncodingNotSupportedException>(delegate {
          testFunc(mailbox,
                   ImapThreadingAlgorithm.References,
                   ImapSearchCriteria.Subject("test"),
                   Encoding.GetEncoding("Shift_JIS"));
        });

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("shift_jis", ex.EncodingName);
      });
    }

    [Test]
    public void TestGetSortedMessages_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.Sort},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Date,
                                                 ImapSearchCriteria.Subject("test"),
                                                 Encoding.GetEncoding("Shift_JIS"));

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SORT (DATE) shift_jis SUBJECT {4}\r\ntest\r\n",
                                       "$tag NO [BADCHARSET] charset not supported\r\n");
        });

        var ex = Assert.Throws<Smdn.Formats.EncodingNotSupportedException>(() => messages.GetEnumerator());

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("shift_jis", ex.EncodingName);
      });
    }

    [Test]
    public void TestGetMessagesGmailSearch()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox(new[] {ImapCapability.GimapGmExt1},
                                   "INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessagesGmailSearch("from:太郎 has:attachment");

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SEARCH CHARSET ",
                                       "* SEARCH 1 2 3\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("FETCH 1:3 (UID)\r\n",
                                       "* FETCH 1 (UID 3)\r\n" +
                                       "* FETCH 2 (UID 5)\r\n" +
                                       "* FETCH 3 (UID 7)\r\n" +
                                       "$tag OK done\r\n");
        });

        using (var enumerator = messages.GetEnumerator()) {
          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);

          StringAssert.EndsWith("SEARCH CHARSET utf-8 X-GM-RAW {26}\r\n" + 
                                "from:太郎 has:attachment\r\n",
                                server.DequeueRequest(Encoding.UTF8));

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessagesGmailSearch_GimapGmExt1Incapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
          "$tag OK [READ-WRITE] done\r\n";

      TestUtils2.TestOpenedMailbox("INBOX",
                                   selectResp,
                                   delegate(ImapPseudoServer2 server, ImapOpenedMailboxInfo mailbox) {
        var ex = Assert.Throws<ImapIncapableException>(() => mailbox.GetMessagesGmailSearch("from:太郎 has:attachment"));

        Assert.IsFalse(mailbox.Client.IsBusy);

        CollectionAssert.Contains(ex.RequiredCapabilities, ImapCapability.GimapGmExt1);
      });
    }
  }
}

