using System;
using NUnit.Framework;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapNamespaceTests {
    [Test]
    public void TestNamespaceDescListProperties_ReadOnly()
    {
#if !NET_4_5
      //(("" "/")) (("Other Users/" "/")) NIL
      var ns = new ImapNamespace(new[] {new ImapNamespaceDesc("", "/")},
                                 new[] {new ImapNamespaceDesc("Other Users/", "/")},
                                 new ImapNamespaceDesc[0]);

      Assert.IsTrue(ns.PersonalNamespaces.IsReadOnly, "PersonalNamespaces");
      Assert.IsTrue(ns.OtherUsersNamespaces.IsReadOnly, "OtherUsersNamespaces");
      Assert.IsTrue(ns.SharedNamespaces.IsReadOnly, "SharedNamespaces");

      Assert.Throws<NotSupportedException>(() => ns.PersonalNamespaces.Clear(), "PersonalNamespaces");
      Assert.Throws<NotSupportedException>(() => ns.OtherUsersNamespaces.Clear(), "OtherUsersNamespaces");
      Assert.Throws<NotSupportedException>(() => ns.SharedNamespaces.Clear(), "SharedNamespaces");
#endif
    }

    [Test]
    public void TestNamespaceDescListProperties_NonNull()
    {
      var ns = new ImapNamespace(null, null, null);

      CollectionAssert.IsEmpty(ns.PersonalNamespaces, "PersonalNamespaces");
      CollectionAssert.IsEmpty(ns.OtherUsersNamespaces, "OtherUsersNamespaces");
      CollectionAssert.IsEmpty(ns.SharedNamespaces, "SharedNamespaces");
    }

    [Test]
    public void TestClone()
    {
      //(("" "/")) (("Other Users/" "/")) NIL
      var ns = new ImapNamespace(new[] {new ImapNamespaceDesc("", "/")},
                                 new[] {new ImapNamespaceDesc("Other Users/", "/")},
                                 new ImapNamespaceDesc[0]);
      var cloned = ns.Clone();

      Assert.AreNotSame(ns, cloned);

      Assert.AreNotSame(ns.PersonalNamespaces, cloned.PersonalNamespaces);
      Assert.AreEqual(ns.PersonalNamespaces.Count, cloned.PersonalNamespaces.Count);
      Assert.AreNotSame(ns.PersonalNamespaces[0], cloned.PersonalNamespaces[0]);
      Assert.AreEqual(ns.PersonalNamespaces[0].Prefix, cloned.PersonalNamespaces[0].Prefix);
      Assert.AreEqual(ns.PersonalNamespaces[0].HierarchyDelimiter, cloned.PersonalNamespaces[0].HierarchyDelimiter);

      Assert.AreNotSame(ns.OtherUsersNamespaces, cloned.OtherUsersNamespaces);
      Assert.AreEqual(ns.OtherUsersNamespaces.Count, cloned.OtherUsersNamespaces.Count);
      Assert.AreNotSame(ns.OtherUsersNamespaces[0], cloned.OtherUsersNamespaces[0]);

      Assert.AreNotSame(ns.SharedNamespaces, cloned.SharedNamespaces);
      Assert.AreEqual(ns.SharedNamespaces.Count, cloned.SharedNamespaces.Count);
    }

    [Test]
    public void TestToString()
    {
      var ns = new ImapNamespace(new[] {new ImapNamespaceDesc("", "/")},
                                 new[] {new ImapNamespaceDesc("Other Users/", "/")},
                                 new ImapNamespaceDesc[0]);

      Assert.DoesNotThrow(() => ns.ToString());
    }
  }
}
