//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2010-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

#if ASYNC_OPS
using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;

namespace Smdn.Net.Pop3.Client {
  partial class PopClient {
#if NET_4_5
    public Task<IReadOnlyList<PopMessageInfo>> GetMessagesAsync()
#else
    public Task<IList<PopMessageInfo>> GetMessagesAsync()
#endif
    {
      return GetMessagesAsync(defaultGetUniqueId,
                              CancellationToken.None);
    }

#if NET_4_5
    public Task<IReadOnlyList<PopMessageInfo>> GetMessagesAsync(CancellationToken cancellationToken)
#else
    public Task<IList<PopMessageInfo>> GetMessagesAsync(CancellationToken cancellationToken)
#endif
    {
      return GetMessagesAsync(defaultGetUniqueId,
                              cancellationToken);
    }

#if NET_4_5
    public Task<IReadOnlyList<PopMessageInfo>> GetMessagesAsync(bool getUniqueId)
#else
    public Task<IList<PopMessageInfo>> GetMessagesAsync(bool getUniqueId)
#endif
    {
      return GetMessagesAsync(getUniqueId,
                              CancellationToken.None);
    }

#if NET_4_5
    public Task<IReadOnlyList<PopMessageInfo>> GetMessagesAsync(bool getUniqueId,
                                                                CancellationToken cancellationToken)
#else
    public Task<IList<PopMessageInfo>> GetMessagesAsync(bool getUniqueId,
                                                        CancellationToken cancellationToken)
#endif
    {
      return StartTaskExclusively(() => GetMessagesCore(getUniqueId),
                                  cancellationToken);
    }

    public Task<PopMessageInfo> GetMessageAsync(long messageNumber)
    {
      return GetMessageAsync(messageNumber,
                             defaultGetUniqueId,
                             CancellationToken.None);
    }

    public Task<PopMessageInfo> GetMessageAsync(long messageNumber,
                                                CancellationToken cancellationToken)
    {
      return GetMessageAsync(messageNumber,
                             defaultGetUniqueId,
                             cancellationToken);
    }

    public Task<PopMessageInfo> GetMessageAsync(long messageNumber,
                                                bool getUniqueId)
    {
      return GetMessageAsync(messageNumber,
                             getUniqueId,
                             CancellationToken.None);
    }

    public Task<PopMessageInfo> GetMessageAsync(long messageNumber,
                                                bool getUniqueId,
                                                CancellationToken cancellationToken)
    {
      if (messageNumber <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("messageNumber", messageNumber);

      return StartTaskExclusively(() => GetMessageByMessageNumber(messageNumber, getUniqueId),
                                  cancellationToken);
    }

    public Task<PopMessageInfo> GetFirstMessageAsync()
    {
      return GetFirstMessageAsync(defaultGetUniqueId,
                                  CancellationToken.None);
    }

    public Task<PopMessageInfo> GetFirstMessageAsync(CancellationToken cancellationToken)
    {
      return GetFirstMessageAsync(defaultGetUniqueId,
                                  cancellationToken);
    }

    public Task<PopMessageInfo> GetFirstMessageAsync(bool getUniqueId)
    {
      return GetFirstMessageAsync(getUniqueId,
                                  CancellationToken.None);
    }

    public Task<PopMessageInfo> GetFirstMessageAsync(bool getUniqueId,
                                                     CancellationToken cancellationToken)
    {
      return GetMessageAsync(1L,
                             getUniqueId,
                             cancellationToken);
    }

    public Task<PopMessageInfo> GetLastMessageAsync()
    {
      return GetLastMessageAsync(defaultGetUniqueId,
                                 CancellationToken.None);
    }

    public Task<PopMessageInfo> GetLastMessageAsync(CancellationToken cancellationToken)
    {
      return GetLastMessageAsync(defaultGetUniqueId,
                                 cancellationToken);
    }

    public Task<PopMessageInfo> GetLastMessageAsync(bool getUniqueId)
    {
      return GetLastMessageAsync(getUniqueId,
                                 CancellationToken.None);
    }

    public Task<PopMessageInfo> GetLastMessageAsync(bool getUniqueId,
                                                    CancellationToken cancellationToken)
    {
      return StartTaskExclusively(() => GetLastMessageCore(getUniqueId),
                                  cancellationToken);
    }

    public Task<PopMessageInfo> GetMessageAsync(string uniqueId)
    {
      return GetMessageAsync(uniqueId,
                             CancellationToken.None);
    }

    public Task<PopMessageInfo> GetMessageAsync(string uniqueId,
                                                CancellationToken cancellationToken)
    {
      uniqueId = GetValidatedUniqueId(uniqueId);

      return StartTaskExclusively(() => GetMessageByUniqueId(uniqueId),
                                  cancellationToken);
    }

    public Task UpdateStatusAsync()
    {
      return StartTaskExclusively(UpdateStatusCore,
                                  CancellationToken.None);
    }

    public Task CancelDeleteAsync()
    {
      return StartTaskExclusively(CancelDeleteCore,
                                  CancellationToken.None);
    }

    public Task KeepAliveAsync()
    {
      return StartTaskExclusively(KeepAliveCore,
                                  CancellationToken.None);
    }
  }
}
#endif // ASYNC_OPS