// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2010-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.Linq;
#if ASYNC_OPS
using System.Threading;
using System.Threading.Tasks;
#endif
using System.IO;
using System.Text;

namespace Smdn.Net.Pop3.Client {
  partial class PopClient {
    public string DownloadMessageAsText(long messageNumber, Encoding encoding)
    {
      return GetMessage(messageNumber).ReadAllText(encoding);
    }

#if ASYNC_OPS
    public Task<string> DownloadMessageAsTextAsync(long messageNumber, Encoding encoding)
    {
      return GetMessageAsync(messageNumber).ContinueWith(t => t.Result.ReadAllText(encoding));
    }
#endif

    public string DownloadFirstMessageAsText(Encoding encoding)
    {
      return GetFirstMessage().ReadAllText(encoding);
    }

#if ASYNC_OPS
    public Task<string> DownloadFirstMessageAsTextAsync(Encoding encoding)
    {
      return GetFirstMessageAsync().ContinueWith(t => t.Result.ReadAllText(encoding));
    }
#endif

    public string DownloadLastMessageAsText(Encoding encoding)
    {
      return GetLastMessage().ReadAllText(encoding);
    }

#if ASYNC_OPS
    public Task<string> DownloadLastMessageAsTextAsync(Encoding encoding)
    {
      return GetLastMessageAsync().ContinueWith(t => t.Result.ReadAllText(encoding));
    }
#endif

    public IEnumerable<string> DownloadAllMessagesAsText(Encoding encoding)
    {
      foreach (var message in GetMessages(false)) {
        yield return message.ReadAllText(encoding);
      }
    }

#if ASYNC_OPS
#if NET_4_5
    public Task<IReadOnlyList<string>> DownloadAllMessagesAsTextAsync(Encoding encoding)
#else
    public Task<IList<string>> DownloadAllMessagesAsTextAsync(Encoding encoding)
#endif
    {
      return DownloadAllMessagesAsyncCore(message => message.ReadAllText(encoding));
    }

#if NET_4_5
    private Task<IReadOnlyList<TResult>> DownloadAllMessagesAsyncCore<TResult>(Func<PopMessageInfo, TResult> selector)
#else
    private Task<IList<TResult>> DownloadAllMessagesAsyncCore<TResult>(Func<PopMessageInfo, TResult> selector)
#endif
    {
      return GetMessagesAsync(false).ContinueWith(t => {
        var ret = t.Result.Select(selector).ToList();

#if NET_4_5
        return (IReadOnlyList<TResult>)ret;
#else
        return (IList<TResult>)ret.AsReadOnly();
#endif
      });
    }
#endif

    public byte[] DownloadMessageAsByteArray(long messageNumber)
    {
      return GetMessage(messageNumber).ReadAllBytes();
    }

#if ASYNC_OPS
    public Task<byte[]> DownloadMessageAsByteArrayAsync(long messageNumber)
    {
      return GetMessageAsync(messageNumber).ContinueWith(t => t.Result.ReadAllBytes());
    }
#endif

    public byte[] DownloadFirstMessageAsByteArray()
    {
      return GetFirstMessage().ReadAllBytes();
    }

#if ASYNC_OPS
    public Task<byte[]> DownloadFirstMessageAsByteArrayAsync()
    {
      return GetFirstMessageAsync().ContinueWith(t => t.Result.ReadAllBytes());
    }
#endif

    public byte[] DownloadLastMessageAsByteArray()
    {
      return GetLastMessage().ReadAllBytes();
    }

#if ASYNC_OPS
    public Task<byte[]> DownloadLastMessageAsByteArrayAsync()
    {
      return GetLastMessageAsync().ContinueWith(t => t.Result.ReadAllBytes());
    }
#endif

    public Stream DownloadMessageAsStream(long messageNumber)
    {
      return GetMessage(messageNumber).OpenRead();
    }

#if ASYNC_OPS
    public Task<Stream> DownloadMessageAsStreamAsync(long messageNumber)
    {
      return GetMessageAsync(messageNumber).ContinueWith(t => t.Result.OpenRead());
    }
#endif

    public Stream DownloadFirstMessageAsStream()
    {
      return GetFirstMessage().OpenRead();
    }

#if ASYNC_OPS
    public Task<Stream> DownloadFirstMessageAsStreamAsync()
    {
      return GetFirstMessageAsync().ContinueWith(t => t.Result.OpenRead());
    }
#endif

    public Stream DownloadLastMessageAsStream()
    {
      return GetLastMessage().OpenRead();
    }

#if ASYNC_OPS
    public Task<Stream> DownloadLastMessageAsStreamAsync()
    {
      return GetLastMessageAsync().ContinueWith(t => t.Result.OpenRead());
    }
#endif

    public TResult DownloadMessageAs<TResult>(long messageNumber, Converter<Stream, TResult> converter)
    {
      return GetMessage(messageNumber).ReadAs(converter);
    }

#if ASYNC_OPS
    public Task<TResult> DownloadMessageAsAsync<TResult>(long messageNumber, Converter<Stream, TResult> converter)
    {
      return GetMessageAsync(messageNumber).ContinueWith(t => t.Result.ReadAs(converter));
    }
#endif

    public TResult DownloadMessageAs<T, TResult>(long messageNumber, Func<Stream, T, TResult> read, T arg)
    {
      return GetMessage(messageNumber).ReadAs(read, arg);
    }

#if ASYNC_OPS
    public Task<TResult> DownloadMessageAsAsync<T, TResult>(long messageNumber, Func<Stream, T, TResult> read, T arg)
    {
      return GetMessageAsync(messageNumber).ContinueWith(t => t.Result.ReadAs(read, arg));
    }
#endif

    public TResult DownloadFirstMessageAs<TResult>(Converter<Stream, TResult> converter)
    {
      return GetFirstMessage().ReadAs(converter);
    }

#if ASYNC_OPS
    public Task<TResult> DownloadFirstMessageAsAsync<TResult>(Converter<Stream, TResult> converter)
    {
      return GetFirstMessageAsync().ContinueWith(t => t.Result.ReadAs(converter));
    }
#endif

    public TResult DownloadFirstMessageAs<T, TResult>(Func<Stream, T, TResult> read, T arg)
    {
      return GetFirstMessage().ReadAs(read, arg);
    }

#if ASYNC_OPS
    public Task<TResult> DownloadFirstMessageAsAsync<T, TResult>(Func<Stream, T, TResult> read, T arg)
    {
      return GetFirstMessageAsync().ContinueWith(t => t.Result.ReadAs(read, arg));
    }
#endif

    public TResult DownloadLastMessageAs<TResult>(Converter<Stream, TResult> converter)
    {
      return GetLastMessage().ReadAs(converter);
    }

#if ASYNC_OPS
    public Task<TResult> DownloadLastMessageAsAsync<TResult>(Converter<Stream, TResult> converter)
    {
      return GetLastMessageAsync().ContinueWith(t => t.Result.ReadAs(converter));
    }
#endif

    public TResult DownloadLastMessageAs<T, TResult>(Func<Stream, T, TResult> read, T arg)
    {
      return GetLastMessage().ReadAs(read, arg);
    }

#if ASYNC_OPS
    public Task<TResult> DownloadLastMessageAsAsync<T, TResult>(Func<Stream, T, TResult> read, T arg)
    {
      return GetLastMessageAsync().ContinueWith(t => t.Result.ReadAs(read, arg));
    }
#endif

    public IEnumerable<TResult> DownloadAllMessagesAs<TResult>(Converter<Stream, TResult> converter)
    {
      foreach (var message in GetMessages(false)) {
        yield return message.ReadAs(converter);
      }
    }

#if ASYNC_OPS
#if NET_4_5
    public Task<IReadOnlyList<TResult>> DownloadAllMessagesAsAsync<TResult>(Converter<Stream, TResult> converter)
#else
    public Task<IList<TResult>> DownloadAllMessagesAsAsync<TResult>(Converter<Stream, TResult> converter)
#endif
    {
      return DownloadAllMessagesAsyncCore(message => message.ReadAs(converter));
    }
#endif

    public IEnumerable<TResult> DownloadAllMessagesAs<T, TResult>(Func<Stream, T, TResult> read, T arg)
    {
      foreach (var message in GetMessages(false)) {
        yield return message.ReadAs(read, arg);
      }
    }

#if ASYNC_OPS
#if NET_4_5
    public Task<IReadOnlyList<TResult>> DownloadAllMessagesAsAsync<T, TResult>(Func<Stream, T, TResult> read, T arg)
#else
    public Task<IList<TResult>> DownloadAllMessagesAsAsync<T, TResult>(Func<Stream, T, TResult> read, T arg)
#endif
    {
      return DownloadAllMessagesAsyncCore(message => message.ReadAs(read, arg));
    }
#endif

    public FileInfo DownloadMessageToFile(long messageNumber, string path)
    {
      var file = new FileInfo(path);

      GetMessage(messageNumber).Save(file, PopMessageInfo.EntireMessage);

      return file;
    }

#if ASYNC_OPS
    public Task<FileInfo> DownloadMessageToFileAsync(long messageNumber, string path)
    {
      return GetMessageAsync(messageNumber).ContinueWith(t => {
        var file = new FileInfo(path);

        t.Result.Save(file, PopMessageInfo.EntireMessage);

        return file;
      });
    }
#endif

    public FileInfo DownloadFirstMessageToFile(string path)
    {
      var file = new FileInfo(path);

      GetFirstMessage().Save(file, PopMessageInfo.EntireMessage);

      return file;
    }

#if ASYNC_OPS
    public Task<FileInfo> DownloadFirstMessageToFileAsync(string path)
    {
      return GetFirstMessageAsync().ContinueWith(t => {
        var file = new FileInfo(path);

        t.Result.Save(file, PopMessageInfo.EntireMessage);

        return file;
      });
    }
#endif

    public FileInfo DownloadLastMessageToFile(string path)
    {
      var file = new FileInfo(path);

      GetLastMessage().Save(file, PopMessageInfo.EntireMessage);

      return file;
    }

#if ASYNC_OPS
    public Task<FileInfo> DownloadLastMessageToFileAsync(string path)
    {
      return GetLastMessageAsync().ContinueWith(t => {
        var file = new FileInfo(path);

        t.Result.Save(file, PopMessageInfo.EntireMessage);

        return file;
      });
    }
#endif

    public FileInfo[] DownloadAllMessagesToDirectory(string directory)
    {
      if (directory == null)
        throw new ArgumentNullException("directory");
      if (directory.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("directory");

      ThrowIfNotConnected();

      Directory.CreateDirectory(directory);

      return DownloadAllMessagesToFiles(delegate(PopMessageInfo arg) {
        return Path.Combine(directory, string.Format("{0}.eml", arg.MessageNumber));
      });
    }

#if ASYNC_OPS
#if NET_4_5
    public Task<IReadOnlyList<FileInfo>> DownloadAllMessagesToDirectoryAsync(string directory)
#else
    public Task<IList<FileInfo>> DownloadAllMessagesToDirectoryAsync(string directory)
#endif
    {
      if (directory == null)
        throw new ArgumentNullException("directory");
      if (directory.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("directory");

      ThrowIfNotConnected();

      Directory.CreateDirectory(directory);

      return DownloadAllMessagesToFilesAsync(delegate(PopMessageInfo arg) {
        return Path.Combine(directory, string.Format("{0}.eml", arg.MessageNumber));
      });
    }
#endif

    public FileInfo[] DownloadAllMessagesToFiles(Func<PopMessageInfo, string> selectFileName)
    {
      if (selectFileName == null)
        throw new ArgumentNullException("selectFileName");

      return GetMessages(false).Select(delegate(PopMessageInfo message) {
        var file = new FileInfo(selectFileName(message));

        message.Save(file, PopMessageInfo.EntireMessage);

        return file;
      }).ToArray();
    }

#if ASYNC_OPS
#if NET_4_5
    public Task<IReadOnlyList<FileInfo>> DownloadAllMessagesToFilesAsync(Func<PopMessageInfo, string> selectFileName)
#else
    public Task<IList<FileInfo>> DownloadAllMessagesToFilesAsync(Func<PopMessageInfo, string> selectFileName)
#endif
    {
      if (selectFileName == null)
        throw new ArgumentNullException("selectFileName");

      return DownloadAllMessagesAsyncCore(message => {
        var file = new FileInfo(selectFileName(message));

        message.Save(file, PopMessageInfo.EntireMessage);

        return file;
      });
    }
#endif
  }
}