//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2010-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_5
using System;
using System.Threading;

namespace Smdn.Net.Pop3.Client {
  public struct PopMessageRetrieveProgressInfo {
    public long RetrievedOctetCount {
      get; private set;
    }

    public long RetrievedLineCount {
      get; private set;
    }

    public long TotalOctetCount {
      get; private set;
    }

    public int ProgressPercentage {
      get { return (int)((RetrievedOctetCount * 100L) / TotalOctetCount); }
    }

    public CancellationToken CancellationToken {
      get; private set;
    }

    public PopMessageRetrieveProgressInfo(long retrievedOctetCount,
                                          long retrievedLineCount,
                                          long totalOctetCount,
                                          CancellationToken cancellationToken)
      : this()
    {
      this.RetrievedOctetCount = retrievedOctetCount;
      this.RetrievedLineCount = retrievedLineCount;
      this.TotalOctetCount = totalOctetCount;
      this.CancellationToken = cancellationToken;
    }

    public override string ToString()
    {
      return string.Format("[PopMessageRetrieveProgressInfo: RetrievedOctetCount={0}, RetrievedLineCount={1}, TotalOctetCount={2}, ProgressPercentage={3}]",
                           RetrievedOctetCount,
                           RetrievedLineCount,
                           TotalOctetCount,
                           ProgressPercentage);
    }
  }
}
#endif
