using System;
using System.Net;
using NUnit.Framework;

using Smdn.Net.Pop3.Protocol;

namespace Smdn.Net.Pop3.Client.Session {
  public abstract class PopSessionTestsBase {
    protected class NullCredential : ICredentialsByHost {
      public NetworkCredential GetCredential(string host, int port, string authType)
      {
        return null;
      }
    }

    protected void Connect(Action<PopSession> action)
    {
      Connect(delegate(PopSession session, PopPseudoServer server) {
        action(session);
      });
    }

    protected void Connect(Action<PopSession, PopPseudoServer> action)
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var session = new PopSession(server.Host, server.Port)) {
          action(session, server);
        }
      }
    }

    protected void Login(Action<PopSession, PopPseudoServer> action)
    {
      Login(null, action);
    }

    protected void Login(PopCapability[] capabilities, Action<PopSession, PopPseudoServer> action)
    {
      Connect(delegate(PopSession session, PopPseudoServer server) {
        if (capabilities != null && 0 < capabilities.Length) {
          var resp = "+OK Capability list follows\r\n" +
                     string.Join("\r\n", (new PopCapabilitySet(capabilities)).ToStringArray()) +
                     "\r\n.\r\n";

          server.EnqueueResponse("CAPA\r\n",
                                 resp);

          Assert.IsTrue((bool)session.Capa());

          server.DequeueRequest(); //CAPA
        }

        server.EnqueueResponse("USER popuser\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS password\r\n",
                               "+OK\r\n");

        var credential = new NetworkCredential("popuser",
                                               "password",
                                               server.Host);

        Assert.IsTrue((bool)session.Login(credential));

        server.DequeueRequest(); // USER
        server.DequeueRequest(); // PASS

        Assert.AreEqual(PopSessionState.Transaction, session.State);

        session.HandlesIncapableAsException = true;

        action(session, server);
      });
    }
  }
}
