using System;
using System.IO;
using System.Collections.Generic;

using Smdn.Formats.Ini;

namespace Smdn.Applications.MRAConnector {
  public static class ProfileStore {
    private static readonly string profileStoreFile = "profiles.ini";

    public static IDictionary<string, Profile> LoadProfiles(out Profile defaultProfile)
    {
      defaultProfile = new Profile();

      var profiles = new Dictionary<string, Profile>();

      if (!File.Exists(profileStoreFile))
        return profiles;

      var document = IniDocument.Load(profileStoreFile);

      foreach (var section in document.Sections) {
        try {
          if (section.IsDefaultSection)
            defaultProfile = GetProfileSection(section);
          else
            profiles.Add(section.Name, GetProfileSection(section));
        }
        catch (Exception ex) {
          // ignore invalid section
          Console.Error.WriteLine(ex);
        }
      }

      return profiles;
    }

    public static void SaveProfiles(Profile defaultProfile, IDictionary<string, Profile> profiles)
    {
      var document = new IniDocument();

      SetProfileSection(document.DefaultSection, defaultProfile);

      foreach (var profile in profiles) {
        SetProfileSection(document[profile.Key], profile.Value);
      }

      document.Save(profileStoreFile);
    }

    private static Profile GetProfileSection(IniSection section)
    {
      var profile = new Profile(section.GetThrowException<Uri>("authority", ConvertUtils.ToUri));

      profile.SslUsage = section.Get<SslUsage>("sslUsage", SslUsage.IfAvailable, EnumUtils.ParseIgnoreCase<SslUsage>);
      profile.AllowInsecureLogin = section.Get<bool>("allowInsecureLogin", false, Convert.ToBoolean);
      profile.AllowPopStandardLogin = section.Get<bool>("allowPopStandardLogin", true, Convert.ToBoolean);
      profile.AllowApopLogin = section.Get<bool>("allowApopLogin",true, Convert.ToBoolean);
      profile.Timeout = section.Get<int>("timeout", -1, Convert.ToInt32);

      return profile;
    }

    private static void SetProfileSection(IniSection section, Profile profile)
    {
      section["authority"] = profile.Authority.AbsoluteUri;
      section["sslUsage"] = Convert.ToString(profile.SslUsage);
      section["allowInsecureLogin"] = Convert.ToString(profile.AllowInsecureLogin);
      section["allowPopStandardLogin"] = Convert.ToString(profile.AllowPopStandardLogin);
      section["allowApopLogin"] = Convert.ToString(profile.AllowApopLogin);
      section["timeout"] = Convert.ToString(profile.Timeout);
    }
  }
}
